from dataclasses import dataclass
from typing import Self


@dataclass(frozen=True)
class Capabilities:
    """Represents the communication capabilities of a Twilio phone number.

    Attributes:
        fax: Whether the phone number supports fax.
        mms: Whether the phone number supports MMS (multimedia messages).
        sms: Whether the phone number supports SMS (text messages).
        voice: Whether the phone number supports voice calls.
    """

    fax: bool
    mms: bool
    sms: bool
    voice: bool

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a Capability instance from a dictionary.

        Args:
            data: Dictionary containing capability data from Twilio API.

        Returns:
            Capability instance with parsed capability flags.
        """
        return cls(
            fax=bool(data["fax"]),
            mms=bool(data["mms"]),
            sms=bool(data["sms"]),
            voice=bool(data["voice"]),
        )

    def to_dict(self) -> dict:
        """Convert the Capability to a dictionary representation.

        Returns:
            Dictionary containing all capability flags.
        """
        return {
            "fax": self.fax,
            "mms": self.mms,
            "sms": self.sms,
            "voice": self.voice,
        }


__exports__ = ()



from datetime import date

import factory

from canvas_sdk.v1.data import (
    PracticeLocation,
    PracticeLocationAddress,
    PracticeLocationContactPoint,
    PracticeLocationSetting,
)
from canvas_sdk.v1.data.common import (
    AddressState,
    AddressType,
    AddressUse,
    ContactPointSystem,
    ContactPointUse,
)


class PracticeLocationSettingFactory(factory.django.DjangoModelFactory[PracticeLocationSetting]):
    """Factory for PracticeLocationSetting."""

    class Meta:
        model = PracticeLocationSetting
        django_get_or_create = ("name",)

    name = factory.Sequence(lambda n: f"location_setting_{n}")
    value = {"random": "json"}


class PracticeLocationFactory(factory.django.DjangoModelFactory[PracticeLocation]):
    """Factory for PracticeLocation."""

    class Meta:
        model = PracticeLocation

    organization = factory.SubFactory(
        "canvas_sdk.test_utils.factories.organization.OrganizationFactory"
    )
    short_name = "Family Doctors of SF"
    full_name = "Family Doctors of San Francisco"
    setting = factory.RelatedFactory(PracticeLocationSettingFactory, "practice_location")
    bill_through_organization = True
    billing_location_name = "Billing Location Name"
    tax_id = "123456789"
    tax_id_type = "E"
    npi_number = "1234567890"
    group_npi_number = "112233445"


class PracticeLocationAddressFactory(factory.django.DjangoModelFactory[PracticeLocationAddress]):
    """Factory for PracticeLocationAddress."""

    class Meta:
        model = PracticeLocationAddress

    practice_location = factory.SubFactory(
        "canvas_sdk.test_utils.factories.PracticeLocationFactory"
    )
    line1 = "1234 Golden Gate Ave."
    line2 = "Suite 456"
    city = "San Francisco"
    district = ""
    state_code = "CA"
    postal_code = "94100"
    use = AddressUse.WORK
    type = AddressType.BOTH
    longitude = -122.476944
    latitude = 37.769722
    start = date(year=2015, month=10, day=10)
    end = None
    country = "USA"
    state = AddressState.ACTIVE


class PracticeLocationContactPointFactory(
    factory.django.DjangoModelFactory[PracticeLocationContactPoint]
):
    """Factory for creating PracticeLocationContactPoint."""

    class Meta:
        model = PracticeLocationContactPoint

    practice_location = factory.SubFactory(PracticeLocationFactory)
    system = ContactPointSystem.PHONE
    value = "8003701416"
    use = ContactPointUse.WORK
    rank = 1



import factory

from canvas_sdk.v1.data import ServiceProvider


class ServiceProviderFactory(factory.django.DjangoModelFactory[ServiceProvider]):
    """Factory for creating ServiceProvider."""

    class Meta:
        model = ServiceProvider

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    business_fax = factory.Faker("phone_number")
    business_phone = factory.Faker("phone_number")
    business_address = factory.Faker("address")
    specialty = factory.Faker("job")
    practice_name = factory.Faker("company")
    notes = factory.Faker("text", max_nb_chars=200)



from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.v1.data import Claim


class _AddClaimComment(_BaseEffect):
    """Effect to add a comment to a Claim."""

    class Meta:
        effect_type = EffectType.ADD_CLAIM_COMMENT

    claim_id: UUID | str
    comment: str

    @property
    def values(self) -> dict[str, Any]:
        """The values for adding a claim comment."""
        return {"claim_id": str(self.claim_id), "comment": self.comment}

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)
        if not Claim.objects.filter(id=self.claim_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    f"Claim with id {self.claim_id} does not exist.",
                    self.claim_id,
                )
            )
        return errors


__exports__ = ()



from canvas_sdk.clients.sendgrid.constants.attachment_disposition import AttachmentDisposition
from canvas_sdk.clients.sendgrid.constants.constants import Constants
from canvas_sdk.clients.sendgrid.constants.criterion_operation import CriterionOperation
from canvas_sdk.clients.sendgrid.constants.event_email import EventEmail
from canvas_sdk.clients.sendgrid.constants.recipient_type import RecipientType
from canvas_sdk.clients.sendgrid.constants.status_email import StatusEmail

__all__ = __exports__ = (
    "AttachmentDisposition",
    "Constants",
    "CriterionOperation",
    "EventEmail",
    "RecipientType",
    "StatusEmail",
)



import importlib.metadata

import deprecation

from canvas_sdk.effects.claim.payment.claim_payment import _PostClaimPayment

version = importlib.metadata.version("canvas")


@deprecation.deprecated(
    deprecated_in="0.95.0",
    removed_in="1.0.0",
    current_version=version,
    details="Use canvas_sdk.effects.claim.ClaimEffect::post_payment instead",
)
class PostClaimPayment(_PostClaimPayment):
    """Deprecated. Instead, use canvas_sdk.effects.claim.ClaimEffect::post_payment."""

    pass


__exports__ = ("PostClaimPayment",)



from dataclasses import dataclass
from typing import Self

from canvas_sdk.clients.extend_ai.constants.processor_type import ProcessorType
from canvas_sdk.clients.extend_ai.constants.run_status import RunStatus
from canvas_sdk.clients.extend_ai.structures.processor_meta import ProcessorMeta
from canvas_sdk.clients.extend_ai.structures.processor_result import (
    ResultClassification,
    ResultExtraction,
    ResultSplitter,
)
from canvas_sdk.clients.extend_ai.structures.stored_file import StoredFile
from canvas_sdk.clients.extend_ai.structures.structure import Structure


@dataclass(frozen=True)
class ProcessorRun(Structure):
    """Represents a single execution of an Extend AI processor.

    Attributes:
        id: The unique identifier for the processor run.
        processor: Metadata about the processor that was executed.
        output: The output data produced by the processor run.
        status: The current status of the processor run.
        files: List of files associated with this processor run.
        usage: Total credit usage for this processor run.
    """

    id: str
    processor: ProcessorMeta
    output: ResultClassification | ResultExtraction | ResultSplitter | None
    status: RunStatus
    files: list[StoredFile]
    usage: int

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a ProcessorRun instance from a dictionary.

        Args:
            data: Dictionary containing processor run data from the API.

        Returns:
            A new ProcessorRun instance with calculated total usage.
        """
        usage = [file.get("usage", {}).get("credits", 0) for file in data["files"]]
        usage.append(data.get("usage", {}).get("credits", 0))

        processor = ProcessorMeta(
            id=data["processorId"],
            name=data["processorName"],
            type=ProcessorType(data["type"]),
            created_at=None,
            updated_at=None,
        )
        status = RunStatus(data["status"])
        output: ResultClassification | ResultSplitter | ResultExtraction | None = None
        if status == RunStatus.PROCESSED:
            if processor.type == ProcessorType.CLASSIFY:
                output = ResultClassification.from_dict(data["output"])
            elif processor.type == ProcessorType.SPLITTER:
                output = ResultSplitter.from_dict(data["output"])
            else:  # if processor.type == ProcessorType.EXTRACT:
                output = ResultExtraction.from_dict(data["output"])

        return cls(
            id=data["id"],
            processor=processor,
            output=output,
            status=status,
            files=[StoredFile.from_dict(item) for item in data["files"]],
            usage=sum(usage),
        )

    def to_dict(self) -> dict:
        """Convert this ProcessorRun to a dictionary.

        Returns:
            Dictionary representation suitable for API requests.
        """
        output = None
        if self.output is not None:
            output = self.output.to_dict()
        return {
            "id": self.id,
            "processor": self.processor.to_dict(),
            "output": output,
            "status": self.status.value,
            "files": [f.to_dict() for f in self.files],
            "usage": self.usage,
        }


__exports__ = ()



from typing import Self

from django.db import models

from canvas_sdk.v1.data.base import BaseQuerySet, IdentifiableModel, Model


class BaseReportTemplateQuerySet(BaseQuerySet):
    """Base QuerySet with shared filtering methods for report templates."""

    def active(self) -> Self:
        """Filter to active templates."""
        return self.filter(active=True)

    def custom(self) -> Self:
        """Filter to custom (user-created) templates."""
        return self.filter(custom=True)

    def builtin(self) -> Self:
        """Filter to built-in (system) templates."""
        return self.filter(custom=False)


class BaseReportTemplate(IdentifiableModel):
    """Abstract base model for report templates."""

    class Meta:
        abstract = True

    name = models.CharField(max_length=250)
    code = models.CharField(max_length=50, blank=True, default="")
    code_system = models.CharField(max_length=50, blank=True, default="")
    search_keywords = models.CharField(max_length=500, blank=True, default="")
    active = models.BooleanField(default=True)
    custom = models.BooleanField(default=True)

    def __str__(self) -> str:
        return self.name


class BaseReportTemplateField(Model):
    """Abstract base model for report template fields."""

    class Meta:
        abstract = True

    sequence = models.IntegerField(default=1)
    code = models.CharField(max_length=50, blank=True, null=True)
    code_system = models.CharField(max_length=50, blank=True, default="")
    label = models.CharField(max_length=250)
    units = models.CharField(max_length=50, blank=True, null=True)
    type = models.CharField(max_length=250)
    required = models.BooleanField(default=False)

    def __str__(self) -> str:
        return self.label


class BaseReportTemplateFieldOption(Model):
    """Abstract base model for report template field options."""

    class Meta:
        abstract = True

    label = models.CharField(max_length=250)
    key = models.CharField(max_length=250)

    def __str__(self) -> str:
        return f"{self.label} ({self.key})"


__exports__ = (
    "BaseReportTemplateQuerySet",
    "BaseReportTemplate",
    "BaseReportTemplateField",
    "BaseReportTemplateFieldOption",
)



from dataclasses import dataclass

from canvas_sdk.clients.sendgrid.constants import RecipientType
from canvas_sdk.clients.sendgrid.structures import Address


@dataclass(frozen=True)
class Recipient:
    """Represents an email recipient with type (TO, CC, BCC)."""

    address: Address
    type: RecipientType


__exports__ = ()



import contextlib
import datetime
import json
import logging
import re
import sys
import traceback
from logging import LogRecord
from types import TracebackType
from typing import Any

import requests
from django.conf import settings


class HttpTransport:
    """
    Send messages to Logstash in V1 format.
    """

    def __init__(self, host: str, **kwargs: Any) -> None:
        self.url = host

        self.session = requests.Session()
        self.session.headers.update({"Content-Type": "application/json"})

    def send(self, events: list[Any], **kwargs: Any) -> None:
        """Send events to Logstash."""
        for event in events:
            try:
                self.session.post(self.url, data=event)
            except (KeyboardInterrupt, SystemExit):
                raise
            except BaseException as e:
                print("Logstash exception", e)

    def close(self) -> None:
        """Close the transport."""
        pass


ExcInfo = tuple[type[BaseException], BaseException, TracebackType | None]


def _figure_out_exc_info(v: Any) -> ExcInfo:
    """
    Depending on the Python version will try to do the smartest thing possible
    to transform *v* into an ``exc_info`` tuple.
    """
    if isinstance(v, BaseException):
        return v.__class__, v, v.__traceback__

    if isinstance(v, tuple):
        return v

    if v:
        return sys.exc_info()  # type: ignore[return-value]

    return v


def _json_default(obj: Any) -> str:
    """
    Coerce everything to strings. All objects representing time get output as
    ISO8601.
    """
    if isinstance(obj, datetime.datetime | datetime.date | datetime.time):
        return obj.isoformat()

    return str(obj)


class LogstashFormatterV1(logging.Formatter):
    """
    A custom formatter to prepare logs to be shipped out to logstash V1 format.
    """

    unwanted_fields = [
        "args",
        "created",
        "filename",
        "funcName",
        "levelno",
        "lineno",
        "module",
        "msecs",
        "name",
        "pathname",
        "process",
        "processName",
        "relativeCreated",
        "request",
        "response",
        "threadName",
    ]

    mappings = {
        "levelname": "log_level",
        "name": "log_name",
    }

    def __init__(self) -> None:
        super().__init__()

        self.defaults = {
            "app": "plugin-runner",
            "customer": settings.CUSTOMER_IDENTIFIER,
            "plugin_context": True,
        }

        self.source_host = settings.HOSTNAME

    @staticmethod
    def get_location(fields: dict) -> str:
        """
        Format the log location in a way that saves some bytes.
        """
        path = "?"

        if "pathname" in fields:
            path = re.sub(r"^/plugin-runner/", "", fields["pathname"])
            path = re.sub(r"^/usr/local/lib/python3.\d\d/dist-packages/", "", path)
            path = re.sub(r"^/venvs/plugin-runner/lib/python3.\d\d/site-packages/", "", path)
        elif "filename" in fields:
            path = fields["filename"]

        return "{}:{}".format(path, fields.get("lineno", "?"))

    @staticmethod
    def get_method(fields: dict) -> str:
        """
        Get the method that originated the log.
        """
        return "{}.{}:{}".format(
            fields.get("name", "?"), fields.get("module", "?"), fields.get("funcName", "?")
        )

    def format(self, record: LogRecord) -> str:
        """
        Format a log record to JSON, if the message is a dict assume an empty
        message and use the dict as additional fields.
        """
        fields = record.__dict__.copy()

        if "msg" in fields and isinstance(fields["msg"], dict):
            msg = fields.pop("msg")

            # if the dict has an "event" key, use that as the message
            # to avoid conflicts with other message types that use event as an object.
            if "event" in msg:
                fields["message"] = msg.pop("event")

            fields.update(msg)
        elif "msg" in fields and "message" not in fields:
            msg = record.getMessage()

            del fields["msg"]

            with contextlib.suppress(KeyError, IndexError, ValueError):
                msg = msg.format(**fields)

            fields["message"] = msg

        if "exc_info" in fields:
            if fields["exc_info"]:
                exc_info = _figure_out_exc_info(fields["exc_info"])
                fields["exception_message"] = str(exc_info[1])
                fields["exception_type"] = f"{exc_info[0].__module__}.{exc_info[0].__qualname__}"
                fields["stack_trace"] = traceback.format_exception(*exc_info)
            del fields["exc_info"]

        if "exc_text" in fields and not fields["exc_text"]:
            del fields["exc_text"]

        now = datetime.datetime.now(datetime.UTC)

        base_log = {
            "@timestamp": now.isoformat(timespec="milliseconds").replace("+00:00", "Z"),
            "@version": 1,
            "location": self.get_location(fields),
            "method": self.get_method(fields),
            "source_host": self.source_host,
        }

        for field in self.unwanted_fields:
            if field in fields:
                del fields[field]

        # Remove None fields (like `params` most of the time) and empty
        # argument lists (like `args` most of the time)
        fields = {key: value for key, value in fields.items() if value is not None and value != []}

        # Rename fields with mapping set
        for old_name, new_name in self.mappings.items():
            if old_name in fields:
                fields[new_name] = fields.pop(old_name)

        base_log.update(fields)

        log_record = self.defaults.copy()
        log_record.update(base_log)

        return json.dumps(log_record, default=_json_default)


class LogstashFormatterECS(logging.Formatter):
    """
    A custom formatter to prepare logs to be shipped out to logstash ECS format.
    """

    def __init__(self) -> None:
        super().__init__()

        self.defaults = {
            "service": {"name": "plugin-runner"},
            "labels": {"customer": settings.CUSTOMER_IDENTIFIER},
            "host": {"name": settings.HOSTNAME},
        }

    def format(self, record: LogRecord) -> str:
        """
        Format a log record to JSON, if the message is a dict assume an empty
        message and use the dict as additional fields.
        """
        fields = record.__dict__.copy()

        if "msg" in fields and isinstance(fields["msg"], dict):
            msg = fields.pop("msg")

            # if the dict has an "event" key, use that as the message
            # to avoid conflicts with other message types that use event as an object.
            if "event" in msg:
                fields["message"] = msg.pop("event")

            fields.update(msg)
        elif "msg" in fields and "message" not in fields:
            msg = record.getMessage()

            del fields["msg"]

            with contextlib.suppress(KeyError, IndexError, ValueError):
                msg = msg.format(**fields)

            fields["message"] = msg

        if "exc_info" in fields:
            if fields["exc_info"]:
                exc_info = _figure_out_exc_info(fields["exc_info"])
                fields["exception_message"] = str(exc_info[1])
                fields["exception_type"] = f"{exc_info[0].__module__}.{exc_info[0].__qualname__}"
                fields["stack_trace"] = traceback.format_exception(*exc_info)
            del fields["exc_info"]

        if "exc_text" in fields and not fields["exc_text"]:
            del fields["exc_text"]

        now = datetime.datetime.now(datetime.UTC)

        log_record = {
            **self.defaults,
            "@timestamp": now.isoformat(timespec="milliseconds").replace("+00:00", "Z"),
            "message": fields.get("message", ""),
            "log": {
                "level": fields.get("levelname", ""),
            },
            **(
                {
                    "error": {
                        "message": fields["exception_message"],
                        "type": fields["exception_type"],
                        "stack_trace": fields["stack_trace"],
                    }
                }
                if "exception_message" in fields
                else {}
            ),
        }

        return json.dumps(log_record, default=_json_default)


__exports__ = ()



import json
from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic_core import InitErrorDetails

from canvas_generated.messages.effects_pb2 import Effect
from canvas_sdk.base import TrackableFieldsModel
from canvas_sdk.v1.data.external_event import ExternalEvent as ExternalEventModel


class ExternalEvent(TrackableFieldsModel):
    """
    Effect to create or update an External Event in Canvas.

    External events represent clinical events from external data sources (ADT feeds, etc.)
    such as admissions, discharges, and transfers.

    Example (create):
        effect = ExternalEvent(
            patient_id="patient-uuid",
            visit_identifier="visit-123",
            message_control_id="msg-456",
            event_type="ADT^A01",
        )
        return effect.create()

    Example (update):
        effect = ExternalEvent(
            external_event_id="existing-event-uuid",
            event_cancelation_datetime=datetime.now(),
        )
        return effect.update()
    """

    class Meta:
        effect_type = "EXTERNAL_EVENT"

    external_event_id: str | UUID | None = None  # For updates
    patient_id: str | None = None
    visit_identifier: str | None = None
    message_control_id: str | None = None
    event_type: str | None = None
    event_datetime: datetime | None = None
    event_cancelation_datetime: datetime | None = None
    message_datetime: datetime | None = None
    information_source: str | None = None
    facility_name: str | None = None
    raw_message: str | None = None

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        """Validate the external event data."""
        errors = super()._get_error_details(method)

        if method == "create":
            if self.external_event_id:
                errors.append(
                    self._create_error_detail(
                        "value",
                        "External event ID should not be set when creating a new external event.",
                        self.external_event_id,
                    )
                )
            if not self.patient_id:
                errors.append(
                    self._create_error_detail(
                        "missing",
                        "Field 'patient_id' is required to create an external event.",
                        self.patient_id,
                    )
                )
            if not self.visit_identifier:
                errors.append(
                    self._create_error_detail(
                        "missing",
                        "Field 'visit_identifier' is required to create an external event.",
                        self.visit_identifier,
                    )
                )
            if not self.message_control_id:
                errors.append(
                    self._create_error_detail(
                        "missing",
                        "Field 'message_control_id' is required to create an external event.",
                        self.message_control_id,
                    )
                )
            if not self.event_type:
                errors.append(
                    self._create_error_detail(
                        "missing",
                        "Field 'event_type' is required to create an external event.",
                        self.event_type,
                    )
                )

        if method == "update":
            if not self.external_event_id:
                errors.append(
                    self._create_error_detail(
                        "missing",
                        "Field 'external_event_id' is required to update an external event.",
                        self.external_event_id,
                    )
                )
            elif not ExternalEventModel.objects.filter(id=self.external_event_id).exists():
                errors.append(
                    self._create_error_detail(
                        "value",
                        f"External event with ID {self.external_event_id} does not exist.",
                        self.external_event_id,
                    )
                )

        return errors

    def create(self) -> Effect:
        """Create a new External Event."""
        self._validate_before_effect("create")

        payload = {"data": self.values}

        return Effect(
            type=f"CREATE_{self.Meta.effect_type}",
            payload=json.dumps(payload),
        )

    def update(self) -> Effect:
        """Update an existing External Event."""
        self._validate_before_effect("update")

        payload = {"data": self.values}

        return Effect(
            type=f"UPDATE_{self.Meta.effect_type}",
            payload=json.dumps(payload),
        )


__exports__ = ("ExternalEvent",)



import factory

from canvas_sdk.test_utils.factories.patient import PatientFactory
from canvas_sdk.v1.data import ExternalEvent, ExternalVisit


class ExternalVisitFactory(factory.django.DjangoModelFactory[ExternalVisit]):
    """Factory for creating an ExternalVisit."""

    class Meta:
        model = ExternalVisit

    patient = factory.SubFactory(PatientFactory)
    visit_identifier = factory.Sequence(lambda n: f"visit-{n}")
    information_source = factory.Faker("company")
    facility_name = factory.Faker("company")


class ExternalEventFactory(factory.django.DjangoModelFactory[ExternalEvent]):
    """Factory for creating an ExternalEvent."""

    class Meta:
        model = ExternalEvent

    external_visit = factory.SubFactory(ExternalVisitFactory)
    patient = factory.LazyAttribute(lambda obj: obj.external_visit.patient)
    message_control_id = factory.Sequence(lambda n: f"msg-{n}")
    message_datetime = factory.Faker("date_time_this_year")
    event_type = "ADT^A01"
    event_datetime = factory.Faker("date_time_this_year")
    event_cancelation_datetime = None
    raw_message = factory.Faker("text", max_nb_chars=200)



import json
from builtins import print as builtin_print
from typing import Any

from requests import Response


class Printer:
    """Class to override Python's default print with json and Requests.response capabilities."""

    def __call__(self, *args: Any) -> None:
        """Default printing."""
        self._default_print(*args)

    @staticmethod
    def response(response: Response, success: bool = True) -> None:
        """Print a response object by getting its json or text response."""
        try:
            message = response.json()
        except json.JSONDecodeError:
            message = response.text

        Printer.json(message=message, success=success, status_code=response.status_code)

    @staticmethod
    def json(message: str | None, success: bool = True, **kwargs: Any) -> None:
        """Print a message in json format regardless of the input."""
        status = {"success": success}
        if message:
            try:
                json_message = json.loads(message)
            except (json.JSONDecodeError, TypeError):
                json_message = message

            status.update({"message": json_message})
        for key, value in kwargs.items():
            status.update({key: value})

        Printer._default_print(json.dumps(status))

    @staticmethod
    def _default_print(*args: Any) -> None:
        builtin_print(*args)


print = Printer()



from dataclasses import dataclass
from datetime import datetime

from canvas_sdk.clients.sendgrid.constants.constants import Constants
from canvas_sdk.clients.sendgrid.constants.criterion_operation import CriterionOperation


@dataclass(frozen=True)
class CriterionDatetime:
    """Represents a datetime comparison criterion for SendGrid queries.

    https://www.twilio.com/docs/sendgrid/api-reference/email-logs/filter-all-messages
    """

    date_time: datetime
    operation: CriterionOperation

    def to_str(self) -> str:
        """Convert datetime criterion to query string format."""
        operation = self.operation.value
        timestamp = self.date_time.strftime(Constants.rfc3339_format)
        return f'sg_message_id_created_at {operation} TIMESTAMP "{timestamp}"'


__exports__ = ()



from canvas_sdk.clients.llms.structures.base_model_llm_json import BaseModelLlmJson
from canvas_sdk.clients.llms.structures.file_content import FileContent
from canvas_sdk.clients.llms.structures.llm_file_url import LlmFileUrl
from canvas_sdk.clients.llms.structures.llm_response import LlmResponse
from canvas_sdk.clients.llms.structures.llm_tokens import LlmTokens
from canvas_sdk.clients.llms.structures.llm_turn import LlmTurn

__all__ = __exports__ = (
    "BaseModelLlmJson",
    "FileContent",
    "LlmResponse",
    "LlmTokens",
    "LlmTurn",
    "LlmFileUrl",
)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from canvas_generated.messages import events_pb2 as canvas__generated_dot_messages_dot_events__pb2
from canvas_generated.messages import plugins_pb2 as canvas__generated_dot_messages_dot_plugins__pb2


class PluginRunnerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.HandleEvent = channel.unary_stream(
                '/canvas.PluginRunner/HandleEvent',
                request_serializer=canvas__generated_dot_messages_dot_events__pb2.Event.SerializeToString,
                response_deserializer=canvas__generated_dot_messages_dot_events__pb2.EventResponse.FromString,
                )
        self.ReloadPlugins = channel.unary_stream(
                '/canvas.PluginRunner/ReloadPlugins',
                request_serializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsRequest.SerializeToString,
                response_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsResponse.FromString,
                )
        self.ReloadPlugin = channel.unary_stream(
                '/canvas.PluginRunner/ReloadPlugin',
                request_serializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginRequest.SerializeToString,
                response_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginResponse.FromString,
                )
        self.UnloadPlugin = channel.unary_stream(
                '/canvas.PluginRunner/UnloadPlugin',
                request_serializer=canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginRequest.SerializeToString,
                response_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginResponse.FromString,
                )


class PluginRunnerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def HandleEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReloadPlugins(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReloadPlugin(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UnloadPlugin(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PluginRunnerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'HandleEvent': grpc.unary_stream_rpc_method_handler(
                    servicer.HandleEvent,
                    request_deserializer=canvas__generated_dot_messages_dot_events__pb2.Event.FromString,
                    response_serializer=canvas__generated_dot_messages_dot_events__pb2.EventResponse.SerializeToString,
            ),
            'ReloadPlugins': grpc.unary_stream_rpc_method_handler(
                    servicer.ReloadPlugins,
                    request_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsRequest.FromString,
                    response_serializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsResponse.SerializeToString,
            ),
            'ReloadPlugin': grpc.unary_stream_rpc_method_handler(
                    servicer.ReloadPlugin,
                    request_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginRequest.FromString,
                    response_serializer=canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginResponse.SerializeToString,
            ),
            'UnloadPlugin': grpc.unary_stream_rpc_method_handler(
                    servicer.UnloadPlugin,
                    request_deserializer=canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginRequest.FromString,
                    response_serializer=canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'canvas.PluginRunner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class PluginRunner(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def HandleEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/canvas.PluginRunner/HandleEvent',
            canvas__generated_dot_messages_dot_events__pb2.Event.SerializeToString,
            canvas__generated_dot_messages_dot_events__pb2.EventResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReloadPlugins(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/canvas.PluginRunner/ReloadPlugins',
            canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsRequest.SerializeToString,
            canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReloadPlugin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/canvas.PluginRunner/ReloadPlugin',
            canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginRequest.SerializeToString,
            canvas__generated_dot_messages_dot_plugins__pb2.ReloadPluginResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UnloadPlugin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/canvas.PluginRunner/UnloadPlugin',
            canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginRequest.SerializeToString,
            canvas__generated_dot_messages_dot_plugins__pb2.UnloadPluginResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from canvas_sdk.effects import Effect
from canvas_sdk.effects.launch_modal import LaunchModalEffect
from canvas_sdk.handlers.application import Application


class MyApplication(Application):
    """An embeddable application that can be registered to Canvas."""

    def on_open(self) -> Effect:
        """Handle the on_open event."""
        # Implement this method to handle the application on_open event.
        return LaunchModalEffect(
            url="https://www.canvasmedical.com/extensions",
            target=LaunchModalEffect.TargetType.DEFAULT_MODAL,
        ).apply()



from canvas_sdk.clients.extend_ai.structures.config.advanced_options_classification import (
    AdvancedOptionsClassification,
)
from canvas_sdk.clients.extend_ai.structures.config.advanced_options_extraction import (
    AdvancedOptionsExtraction,
)
from canvas_sdk.clients.extend_ai.structures.config.advanced_options_splitter import (
    AdvancedOptionsSplitter,
)
from canvas_sdk.clients.extend_ai.structures.config.config_classification import (
    ConfigClassification,
)
from canvas_sdk.clients.extend_ai.structures.config.config_extraction import ConfigExtraction
from canvas_sdk.clients.extend_ai.structures.config.config_splitter import ConfigSplitter
from canvas_sdk.clients.extend_ai.structures.config.parser import Parser
from canvas_sdk.clients.extend_ai.structures.config.part_range import PartRange

__all__ = __exports__ = (
    "AdvancedOptionsClassification",
    "AdvancedOptionsExtraction",
    "AdvancedOptionsSplitter",
    "ConfigClassification",
    "ConfigExtraction",
    "ConfigSplitter",
    "Parser",
    "PartRange",
)



from django.db import models
from django.db.models import TextChoices

from canvas_sdk.v1.data.base import IdentifiableModel


class CompoundMedication(IdentifiableModel):
    """CompoundMedication."""

    class Meta:
        db_table = "canvas_sdk_data_api_compoundmedication_001"

    class PotencyUnits(TextChoices):
        """Potency Units."""

        APPLICATOR = "C62412", "Applicator"
        BLISTER = "C54564", "Blister"
        CAPLET = "C64696", "Caplet"
        CAPSULE = "C48480", "Capsule"
        EACH = "C64933", "Each"
        FILM = "C53499", "Film"
        GRAM = "C48155", "Gram"
        GUM = "C69124", "Gum"
        IMPLANT = "C48499", "Implant"
        INSERT = "C62276", "Insert"
        KIT = "C48504", "Kit"
        LANCET = "C120263", "Lancet"
        LOZENGE = "C48506", "Lozenge"
        MILLILITER = "C28254", "Milliliter"
        PACKET = "C48521", "Packet"
        PAD = "C65032", "Pad"
        PATCH = "C48524", "Patch"
        PEN_NEEDLE = "C120216", "Pen Needle"
        RING = "C62609", "Ring"
        SPONGE = "C53502", "Sponge"
        STICK = "C53503", "Stick"
        STRIP = "C48538", "Strip"
        SUPPOSITORY = "C48539", "Suppository"
        SWAB = "C53504", "Swab"
        TABLET = "C48542", "Tablet"
        TROCHE = "C48548", "Troche"
        UNSPECIFIED = "C38046", "Unspecified"
        WAFER = "C48552", "Wafer"

    class ControlledSubstanceOptions(TextChoices):
        """Controlled Substance Options."""

        SCHEDULE_NOT_SCHEDULED = "N", "None"
        SCHEDULE_II = "II", "Schedule II"
        SCHEDULE_III = "III", "Schedule III"
        SCHEDULE_IV = "IV", "Schedule IV"
        SCHEDULE_V = "V", "Schedule V"

    active = models.BooleanField(default=True)
    formulation = models.CharField(max_length=105)
    potency_unit_code = models.CharField(max_length=20, choices=PotencyUnits.choices)
    controlled_substance = models.CharField(
        max_length=3, choices=ControlledSubstanceOptions.choices
    )
    controlled_substance_ndc = models.CharField(max_length=20, blank=True, default="")

    def __str__(self) -> str:
        return f'CompoundMedication: "{self.formulation}"'


__exports__ = ("CompoundMedication",)



from typing import Self, cast

from django.db import models

from canvas_sdk.v1.data.base import (
    BaseModelManager,
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
    IdentifiableModel,
    ValueSetLookupQuerySet,
)
from canvas_sdk.v1.data.coding import Coding


class ImmunizationStatus(models.TextChoices):
    """ImmunizationStatus."""

    STATUS_IN_PROGRESS = "in-progress", "In Progress"
    STATUS_ONHOLD = "on-hold"
    STATUS_COMPLETED = "completed"
    STATUS_STOPPED = "stopped"


class ImmunizationReasonsNotGiven(models.TextChoices):
    """ImmunizationReasonsNotGiven."""

    NA = "NA", "not applicable"
    IMMUNE = "IMMUNE", "immunity"
    MEDPREC = "MEDPREC", "medical precaution"
    OSTOCK = "OSTOCK", "product out of stock"
    PATOBJ = "PATOBJ", "patient objection"


class ImmunizationQuerySet(
    CommittableQuerySetMixin,
    ForPatientQuerySetMixin,
    ValueSetLookupQuerySet,
):
    """ImmunizationQuerySet."""

    def active(self) -> Self:
        """Filter immunizations."""
        return self.committed()


ImmunizationManager = BaseModelManager.from_queryset(ImmunizationQuerySet)


class Immunization(IdentifiableModel):
    """Immunization."""

    class Meta:
        db_table = "canvas_sdk_data_api_immunization_001"

    objects = cast(ImmunizationQuerySet, ImmunizationManager())

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="immunizations", null=True
    )
    note = models.ForeignKey("v1.Note", on_delete=models.DO_NOTHING, related_name="immunizations")
    status = models.CharField(
        choices=ImmunizationStatus.choices,
        max_length=20,
        default=ImmunizationStatus.STATUS_IN_PROGRESS,
    )
    lot_number = models.CharField(max_length=20, blank=True, default="")
    manufacturer = models.CharField(max_length=100, blank=True, default="")
    exp_date_original = models.CharField(max_length=50, blank=True, default="")
    exp_date = models.DateField(null=True)
    sig_original = models.CharField(max_length=75, blank=True, default="")
    date_ordered = models.DateField(null=True)
    given_by = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="immunizations_given", null=True
    )
    consent_given = models.BooleanField(default=False)
    take_quantity = models.FloatField(null=True)
    dose_form = models.CharField(max_length=255, blank=True, default="")
    route = models.CharField(max_length=255, blank=True, default="")
    frequency_normalized_per_day = models.FloatField(null=True)
    deleted = models.BooleanField()


class ImmunizationCoding(Coding):
    """ImmunizationCoding."""

    class Meta:
        db_table = "canvas_sdk_data_api_immunizationcoding_001"

    immunization = models.ForeignKey(
        Immunization, on_delete=models.DO_NOTHING, related_name="codings", null=True
    )


class ImmunizationStatementQuerySet(
    CommittableQuerySetMixin, ForPatientQuerySetMixin, ValueSetLookupQuerySet
):
    """ImmunizationStatementQuerySet."""

    def active(self) -> Self:
        """Filter immunizations statements."""
        return self.committed()


ImmunizationStatementManager = BaseModelManager.from_queryset(ImmunizationStatementQuerySet)


class ImmunizationStatement(IdentifiableModel):
    """ImmunizationStatement."""

    class Meta:
        db_table = "canvas_sdk_data_api_immunizationstatement_001"

    objects = cast(ImmunizationStatementQuerySet, ImmunizationStatementManager())

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="immunization_statements", null=True
    )
    note = models.ForeignKey(
        "v1.Note", on_delete=models.DO_NOTHING, related_name="immunization_statements"
    )
    date_original = models.CharField(max_length=50, blank=True, default="")
    date = models.DateField(null=True)
    evidence = models.CharField(max_length=255, blank=True, default="")
    comment = models.CharField(max_length=255, blank=True, default="")
    reason_not_given = models.CharField(
        max_length=20,
        choices=ImmunizationReasonsNotGiven.choices,
        default=ImmunizationReasonsNotGiven.NA,
    )
    deleted = models.BooleanField()


class ImmunizationStatementCoding(Coding):
    """ImmunizationStatementCoding."""

    class Meta:
        db_table = "canvas_sdk_data_api_immunizationstatementcoding_001"

    immunization_statement = models.ForeignKey(
        ImmunizationStatement, on_delete=models.CASCADE, related_name="coding"
    )


__exports__ = (
    "ImmunizationStatus",
    "ImmunizationReasonsNotGiven",
    "Immunization",
    "ImmunizationCoding",
    "ImmunizationStatement",
    "ImmunizationStatementCoding",
)



from dataclasses import dataclass


@dataclass(frozen=True)
class ParsedHeader:
    """Represents a single email header from an inbound parsed email."""

    name: str
    value: str

    def to_dict(self) -> dict:
        """Convert parsed header to dictionary format."""
        return {
            "name": self.name,
            "value": self.value,
        }


__exports__ = ()



import base64
import json
import os
import pathlib
import pickle
import pkgutil
import sys
import threading
from collections import defaultdict
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from http import HTTPStatus
from time import sleep
from typing import Any, TypedDict, cast

import grpc
import redis
import sentry_sdk
from django.core.signals import request_finished, request_started
from redis.backoff import ExponentialBackoff
from redis.exceptions import ConnectionError, TimeoutError
from redis.retry import Retry
from sentry_sdk.integrations.logging import ignore_logger

import settings
from canvas_generated.messages.effects_pb2 import EffectType
from canvas_generated.messages.plugins_pb2 import (
    ReloadPluginRequest,
    ReloadPluginResponse,
    ReloadPluginsRequest,
    ReloadPluginsResponse,
    UnloadPluginRequest,
    UnloadPluginResponse,
)
from canvas_generated.services.plugin_runner_pb2_grpc import (
    PluginRunnerServicer,
    add_PluginRunnerServicer_to_server,
)
from canvas_sdk.effects import Effect
from canvas_sdk.effects.simple_api import Response
from canvas_sdk.events import Event, EventRequest, EventResponse, EventType
from canvas_sdk.handlers.simple_api.websocket import DenyConnection
from canvas_sdk.protocols import ClinicalQualityMeasure
from canvas_sdk.templates.utils import _engine_for_plugin
from canvas_sdk.utils import metrics
from canvas_sdk.utils.metrics import measured
from logger import log
from plugin_runner.authentication import token_for_plugin
from plugin_runner.exceptions import PluginInstallationError, PluginUninstallationError
from plugin_runner.installation import (
    enabled_plugins,
    install_plugin,
    install_plugins,
    uninstall_plugin,
)
from plugin_runner.sandbox import Sandbox, sandbox_from_module
from settings import (
    CHANNEL_NAME,
    CUSTOMER_IDENTIFIER,
    ENV,
    IS_PRODUCTION_CUSTOMER,
    MANIFEST_FILE_NAME,
    PLUGIN_DIRECTORY,
    REDIS_ENDPOINT,
    SECRETS_FILE_NAME,
    SENTRY_DSN,
)

if SENTRY_DSN:
    # Lazy import for faster reload time in dev
    from sentry_sdk.integrations.executing import ExecutingIntegration
    from sentry_sdk.integrations.pure_eval import PureEvalIntegration

    sentry_sdk.init(
        dsn=SENTRY_DSN,
        environment=ENV,
        integrations=[
            ExecutingIntegration(),
            PureEvalIntegration(),
        ],
        release=os.getenv("CANVAS_PLUGINS_REPO_VERSION", "unknown"),
        send_default_pii=True,
        spotlight=False,
        traces_sample_rate=0.0,
        profiles_sample_rate=0.0,
    )

    # Sentry creates an issue for anything logged with logger.error();
    # we want the exceptions themselves, not these error lines
    ignore_logger("plugin_runner_logger")

    global_scope = sentry_sdk.get_global_scope()

    global_scope.set_tag("customer", CUSTOMER_IDENTIFIER)
    global_scope.set_tag("logger", "python")
    global_scope.set_tag("source", "plugin-runner")
    global_scope.set_tag("production_customer", "yes" if IS_PRODUCTION_CUSTOMER else "no")

Plugin = TypedDict(
    "Plugin",
    {
        "active": bool,
        "class": Any,
        "sandbox": Any,
        "handler": Any,
        "secrets": dict[str, str],
    },
)

# a global dictionary of loaded plugins
LOADED_PLUGINS: dict[str, Plugin] = {}

# a global dictionary of values made available to all plugins
ENVIRONMENT: dict = {
    "CUSTOMER_IDENTIFIER": CUSTOMER_IDENTIFIER,
}

# a global dictionary of events to handler class names
EVENT_HANDLER_MAP: dict[str, list] = defaultdict(list)


class DataAccess(TypedDict):
    """DataAccess."""

    event: str
    read: list[str]
    write: list[str]


Protocol = TypedDict(
    "Protocol",
    {
        "class": str,
        "data_access": DataAccess,
    },
)


ApplicationConfig = TypedDict(
    "ApplicationConfig",
    {
        "class": str,
        "description": str,
        "icon": str,
        "scope": str,
    },
)


class Components(TypedDict):
    """Components."""

    protocols: list[Protocol]
    commands: list[dict]
    content: list[dict]
    effects: list[dict]
    views: list[dict]
    applications: list[ApplicationConfig]


class PluginManifest(TypedDict):
    """PluginManifest."""

    sdk_version: str
    plugin_version: str
    name: str
    description: str
    components: Components
    secrets: list[dict]
    tags: dict[str, str]
    references: list[str]
    license: str
    diagram: bool
    readme: str


class PluginRunner(PluginRunnerServicer):
    """This process runs provided plugins that register interest in incoming events."""

    sandbox: Sandbox

    def HandleEvent(self, request: EventRequest, context: Any) -> Iterable[EventResponse]:
        """This is invoked when an event comes in."""
        event = Event(request)
        with metrics.measure(
            metrics.get_qualified_name(self.HandleEvent), extra_tags={"event": event.name}
        ):
            event_type = event.type
            event_name = event.name
            relevant_plugins = EVENT_HANDLER_MAP[event_name]
            relevant_plugin_handlers = []

            log.debug(f"Processing {relevant_plugins} for {event_name}")
            sentry_sdk.set_tag("event-name", event_name)

            if relevant_plugins:
                # Send the Django request_started signal
                request_started.send(sender=self.__class__)

            if event_type in [EventType.PLUGIN_CREATED, EventType.PLUGIN_UPDATED]:
                plugin_name = event.target.id
                # filter only for the plugin(s) that were created/updated
                relevant_plugins = [p for p in relevant_plugins if p.startswith(f"{plugin_name}:")]
            elif event_type in {
                EventType.SIMPLE_API_AUTHENTICATE,
                EventType.SIMPLE_API_REQUEST,
                EventType.SIMPLE_API_WEBSOCKET_AUTHENTICATE,
            }:
                # The target plugin's name will be part of the home-app URL path, so other plugins that
                # respond to SimpleAPI request events are not relevant
                plugin_name = event.context["plugin_name"]
                relevant_plugins = [p for p in relevant_plugins if p.startswith(f"{plugin_name}:")]
            elif event_type in {
                EventType.REVENUE__PAYMENT_PROCESSOR__CHARGE,
                EventType.REVENUE__PAYMENT_PROCESSOR__SELECTED,
                EventType.REVENUE__PAYMENT_PROCESSOR__PAYMENT_METHODS__LIST,
                EventType.REVENUE__PAYMENT_PROCESSOR__PAYMENT_METHODS__ADD,
                EventType.REVENUE__PAYMENT_PROCESSOR__PAYMENT_METHODS__REMOVE,
            }:
                # The target plugin's name will be part of the payment processor identifier, so other plugins that
                # respond to payment processor charge events are not relevant
                try:
                    plugin_name = (
                        base64.b64decode(event.context["identifier"]).decode("utf-8").split(".")[0]
                    )
                    relevant_plugins = [
                        p for p in relevant_plugins if p.startswith(f"{plugin_name}:")
                    ]
                except Exception as ex:
                    log.exception(
                        f"Failed to decode identifier for event {event_name} with context {event.context}"
                    )
                    sentry_sdk.capture_exception(ex)
                    relevant_plugins = []

            effect_list = []

            for plugin_name in relevant_plugins:
                log.debug(f"Processing {plugin_name}")
                sentry_sdk.set_tag("plugin-name", plugin_name)

                plugin = LOADED_PLUGINS[plugin_name]
                handler_class = plugin["class"]
                base_plugin_name = plugin_name.split(":")[0]

                secrets = plugin.get("secrets", {})

                secrets.update(
                    {"graphql_jwt": token_for_plugin(plugin_name=plugin_name, audience="home")}
                )

                try:
                    handler = handler_class(event, secrets, ENVIRONMENT)

                    if not handler.accept_event():
                        continue
                    relevant_plugin_handlers.append(handler_class)

                    classname = (
                        handler.__class__.__name__
                        if isinstance(handler, ClinicalQualityMeasure)
                        else None
                    )
                    handler_name = metrics.get_qualified_name(handler.compute)
                    with metrics.measure(
                        name=handler_name,
                        track_queries=True,
                        track_memory_usage=True,
                        extra_tags={
                            "plugin": base_plugin_name,
                            "event": event_name,
                        },
                    ):
                        _effects = handler.compute()
                        effects = [
                            Effect(
                                type=effect.type,
                                payload=effect.payload,
                                plugin_name=base_plugin_name,
                                classname=classname,
                                handler_name=handler_name,
                                actor=event.actor.id,
                                source=event.source,
                            )
                            for effect in _effects
                        ]

                        effects = validate_effects(effects)

                        apply_effects_to_context(effects, event=event)

                        log.info(f"{plugin_name}.compute() completed.")

                except Exception as e:
                    log.exception(f"Encountered exception in plugin {plugin_name}")
                    sentry_sdk.capture_exception(e)
                    continue

                effect_list += effects

            sentry_sdk.set_tag("plugin-name", None)

            # Special handling for SimpleAPI requests: if there were no relevant handlers (as determined
            # by calling ignore_event on handlers), then set the effects list to be a single 404 Not
            # Found response effect. If multiple handlers were able to respond, log an error and set the
            # effects list to be a single 500 Internal Server Error response effect.
            if event.type in {EventType.SIMPLE_API_AUTHENTICATE, EventType.SIMPLE_API_REQUEST}:
                if len(relevant_plugin_handlers) == 0:
                    effect_list = [Response(status_code=HTTPStatus.NOT_FOUND).apply()]
                elif len(relevant_plugin_handlers) > 1:
                    log.error(
                        f"Multiple handlers responded to {EventType.Name(EventType.SIMPLE_API_REQUEST)}"
                        f" {event.context['path']}"
                    )
                    effect_list = [Response(status_code=HTTPStatus.INTERNAL_SERVER_ERROR).apply()]
            if event.type == EventType.SIMPLE_API_WEBSOCKET_AUTHENTICATE:
                if len(relevant_plugin_handlers) == 0:
                    effect_list = [DenyConnection().apply()]
                elif len(relevant_plugin_handlers) > 1:
                    log.error(
                        f"Multiple handlers responded to {EventType.Name(EventType.SIMPLE_API_WEBSOCKET_AUTHENTICATE)}"
                        f" {event.context['channel']}"
                    )
                    effect_list = [DenyConnection().apply()]

            # Don't log anything if a plugin handler didn't actually run.
            if relevant_plugins:
                # Send the Django request_finished signal
                request_finished.send(sender=self.__class__)

                log.info(f"Responded to Event {event_name}.")

            yield EventResponse(success=True, effects=effect_list)

    def ReloadPlugins(
        self, request: ReloadPluginsRequest, context: Any
    ) -> Iterable[ReloadPluginsResponse]:
        """This is invoked when we need to reload plugins."""
        log.info("Reloading all plugins...")

        message = {"action": "reload"}

        try:
            publish_message(message=message)
        except ImportError:
            yield ReloadPluginsResponse(success=False)
        else:
            yield ReloadPluginsResponse(success=True)

    def ReloadPlugin(
        self, request: ReloadPluginRequest, context: Any
    ) -> Iterable[ReloadPluginResponse]:
        """This is invoked when we need to reload a specific plugin."""
        log.info(f'Reloading plugin "{request.plugin}"...')

        message = {
            "action": "reload",
            "plugin": request.plugin,
        }
        try:
            publish_message(message=message)
        except ImportError:
            yield ReloadPluginResponse(success=False)
        else:
            yield ReloadPluginResponse(success=True)

    def UnloadPlugin(
        self, request: UnloadPluginRequest, context: Any
    ) -> Iterable[UnloadPluginResponse]:
        """This is invoked when we need to reload a specific plugin."""
        log.info(f'Unloading plugin "{request.plugin}"...')

        message = {
            "action": "unload",
            "plugin": request.plugin,
        }
        try:
            publish_message(message=message)
        except ImportError:
            yield UnloadPluginResponse(success=False)
        else:
            yield UnloadPluginResponse(success=True)


STOP_SYNCHRONIZER = threading.Event()


def synchronize_plugins(run_once: bool = False) -> None:
    """
    Listen for messages on the pubsub channel that will indicate it is
    necessary to reinstall and reload plugins.
    """
    log.info(f'synchronize_plugins: listening for messages on pubsub channel "{CHANNEL_NAME}"')

    _, pubsub = get_client()

    pubsub.psubscribe(CHANNEL_NAME)

    while not STOP_SYNCHRONIZER.is_set():
        message = pubsub.get_message(ignore_subscribe_messages=True, timeout=5.0)

        if message is None:
            continue

        log.info(f'synchronize_plugins: received message from pubsub channel "{CHANNEL_NAME}"')

        message_type = message.get("type", "")

        if message_type != "pmessage":
            continue

        data = pickle.loads(message.get("data", pickle.dumps({})))

        if "action" not in data:
            continue

        # clear the template engine cache so that any template changes
        # from plugins are picked up
        _engine_for_plugin.cache_clear()
        plugin_name = data.get("plugin", None)
        try:
            if data["action"] == "reload":
                if plugin_name:
                    plugin = enabled_plugins([plugin_name]).get(plugin_name, None)

                    if plugin:
                        log.info(
                            f'synchronize_plugins: installing/reloading plugin "{plugin_name}" for action=reload'
                        )
                        unload_plugin(plugin_name)
                        install_plugin(plugin_name, attributes=plugin)
                        plugin_dir = pathlib.Path(PLUGIN_DIRECTORY) / plugin_name
                        load_plugin(plugin_dir.resolve())
                else:
                    log.info("synchronize_plugins: installing/reloading plugins for action=reload")
                    install_plugins()
                    load_plugins()
            elif data["action"] == "unload" and plugin_name:
                log.info(f'synchronize_plugins: uninstalling plugin "{plugin_name}"')
                unload_plugin(plugin_name)
                uninstall_plugin(plugin_name)
        except Exception as e:
            if isinstance(e, PluginInstallationError):
                message = "install_plugins failed"
            elif isinstance(e, PluginUninstallationError):
                message = "uninstall_plugin failed"
            else:
                message = "load_plugins failed"

            if plugin_name:
                message += f' for plugin "{plugin_name}"'

            log.exception(f"synchronize_plugins: {message}")
            sentry_sdk.capture_exception(e)

        if run_once:
            break


def synchronize_plugins_and_report_errors() -> None:
    """
    Run synchronize_plugins() in perpetuity and report any encountered errors.
    """
    log.info("synchronize_plugins: starting loop...")

    while not STOP_SYNCHRONIZER.is_set():
        try:
            synchronize_plugins()
        except Exception as e:
            log.exception("synchronize_plugins: error")
            sentry_sdk.capture_exception(e)

        # don't crush redis if we're retrying in a tight loop
        sleep(0.5)


def validate_effects(effects: list[Effect]) -> list[Effect]:
    """
    Validates the effects based on predefined rules.

    Keeps only the first AUTOCOMPLETE_SEARCH_RESULTS effect and preserve all
    non-search-related effects.
    """
    seen_autocomplete = False
    validated_effects = []

    for effect in effects:
        if effect.type == EffectType.AUTOCOMPLETE_SEARCH_RESULTS:
            if seen_autocomplete:
                log.warning("Discarding additional AUTOCOMPLETE_SEARCH_RESULTS effect.")
                continue

            seen_autocomplete = True

        validated_effects.append(effect)

    return validated_effects


def apply_effects_to_context(effects: list[Effect], event: Event) -> Event:
    """Applies AUTOCOMPLETE_SEARCH_RESULTS effects to the event context.
    If we are dealing with a search event, we need to update the context with the search results.
    """
    event_name = event.name

    # Skip if the event is not a search event
    if not event_name.endswith("__PRE_SEARCH") and not event_name.endswith("__POST_SEARCH"):
        return event

    for effect in effects:
        if effect.type == EffectType.AUTOCOMPLETE_SEARCH_RESULTS:
            event.context["results"] = json.loads(effect.payload)
            # Stop processing effects if we've found a AUTOCOMPLETE_SEARCH_RESULTS
            break

    return event


def find_modules(base_path: pathlib.Path, prefix: str | None = None) -> list[str]:
    """Find all modules in the specified package path."""
    modules: list[str] = []

    for _, module_name, is_pkg in pkgutil.iter_modules(
        [base_path.as_posix()],
    ):
        if is_pkg:
            modules = modules + find_modules(
                base_path / module_name,
                prefix=f"{prefix}.{module_name}" if prefix else module_name,
            )
        else:
            modules.append(f"{prefix}.{module_name}" if prefix else module_name)

    return modules


def publish_message(message: dict) -> None:
    """Publish a message to the pubsub channel."""
    log.info(f'Publishing message to pubsub channel "{CHANNEL_NAME}"')
    client, _ = get_client()

    client.publish(CHANNEL_NAME, pickle.dumps(message))


def get_client() -> tuple[redis.Redis, redis.client.PubSub]:
    """Return a Redis client and pubsub object."""
    client = redis.from_url(
        REDIS_ENDPOINT,
        retry=Retry(backoff=ExponentialBackoff(), retries=10),
        retry_on_error=[ConnectionError, TimeoutError, ConnectionResetError],
        health_check_interval=1,
    )
    pubsub = client.pubsub()

    return client, pubsub


@measured(track_memory_usage=True)
def load_or_reload_plugin(path: pathlib.Path) -> bool:
    """Given a path, load or reload a plugin."""
    log.info(f'Loading plugin at "{path}"')

    # the name is the folder name underneath the plugins directory
    name = path.name

    manifest_file = path / MANIFEST_FILE_NAME

    # If installed via `canvas install` we can rely on the manifest file
    # existing. If installed via another method we still need to avoid crashing
    # the entire runner if there's no manifest.
    if not manifest_file.exists():
        log.exception(f'Unable to load plugin "{name}", missing {MANIFEST_FILE_NAME}')
        return False

    manifest_json_str = manifest_file.read_text()

    try:
        manifest_json: PluginManifest = json.loads(manifest_json_str)
    except Exception as e:
        log.exception(f'Unable to load plugin "{name}"')
        sentry_sdk.capture_exception(e)

        return False

    secrets_file = path / SECRETS_FILE_NAME
    secrets_json = {}

    if secrets_file.exists():
        try:
            secrets_json = json.load(secrets_file.open())
        except Exception as e:
            log.exception(f'Unable to load secrets for plugin "{name}"')
            sentry_sdk.capture_exception(e)

    # TODO add existing schema validation from Michela here
    try:
        components = manifest_json["components"]
        handlers = (
            cast(list, components.get("protocols", []))
            + cast(list, components.get("applications", []))
            + cast(list, components.get("handlers", []))
        )
    except Exception as e:
        log.exception(f'Unable to load plugin "{name}"')
        sentry_sdk.capture_exception(e)

        return False

    any_failed = False

    for handler in handlers:
        # TODO add class colon validation to existing schema validation
        # TODO when we encounter an exception here, disable the plugin in response
        try:
            handler_module, handler_class = handler["class"].split(":")
            name_and_class = f"{name}:{handler_module}:{handler_class}"
        except ValueError as e:
            log.exception(f'Unable to parse class for plugin "{name}": "{handler["class"]}"')
            sentry_sdk.capture_exception(e)

            any_failed = True

            continue

        try:
            sandbox = sandbox_from_module(path.parent, handler_module)
            result = sandbox.execute()

            if name_and_class in LOADED_PLUGINS:
                log.info(f"Reloading handler '{name_and_class}'")

                LOADED_PLUGINS[name_and_class]["active"] = True

                LOADED_PLUGINS[name_and_class]["class"] = result[handler_class]
                LOADED_PLUGINS[name_and_class]["sandbox"] = result
                LOADED_PLUGINS[name_and_class]["secrets"] = secrets_json
            else:
                log.info(f'Loading handler "{name_and_class}"')

                LOADED_PLUGINS[name_and_class] = {
                    "active": True,
                    "class": result[handler_class],
                    "sandbox": result,
                    "handler": handler,
                    "secrets": secrets_json,
                }
        except Exception as e:
            log.exception(f"Error importing module '{name_and_class}'")
            sentry_sdk.capture_exception(e)
            any_failed = True

    return not any_failed


def unload_plugin(name: str) -> None:
    """Unload a plugin by its name."""
    handlers_removed = False

    for handler_name in LOADED_PLUGINS.copy():
        if handler_name.startswith(f"{name}:"):
            log.info(f'Unloading handler "{handler_name}"')
            del LOADED_PLUGINS[handler_name]
            handlers_removed = True

    if handlers_removed:
        # Refresh the event type map to remove any handlers for the unloaded plugin
        refresh_event_type_map()
    else:
        log.warning(f"No handlers found for plugin '{name}' to unload.")


def refresh_event_type_map() -> None:
    """Ensure the event subscriptions are up to date."""
    EVENT_HANDLER_MAP.clear()

    for name, plugin in LOADED_PLUGINS.items():
        if hasattr(plugin["class"], "RESPONDS_TO"):
            responds_to = plugin["class"].RESPONDS_TO

            if isinstance(responds_to, str):
                EVENT_HANDLER_MAP[responds_to].append(name)
            elif isinstance(responds_to, list):
                for event in responds_to:
                    EVENT_HANDLER_MAP[event].append(name)
            else:
                log.warning(f"Unknown RESPONDS_TO type: {type(responds_to)}")


@measured
def load_plugins(specified_plugin_paths: list[str] | None = None) -> None:
    """Load the plugins."""
    # first mark each plugin as inactive since we want to remove it from
    # LOADED_PLUGINS if it no longer exists on disk
    for plugin in LOADED_PLUGINS.values():
        plugin["active"] = False

    if specified_plugin_paths is not None:
        # convert to Paths
        plugin_paths = [pathlib.Path(name) for name in specified_plugin_paths]

        for plugin_path in plugin_paths:
            # when we import plugins we'll use the module name directly so we need to add the plugin
            # directory to the path
            path_to_append = pathlib.Path(".") / plugin_path.parent
            sys.path.append(path_to_append.as_posix())
    else:
        # Add plugin directory to path only when actually loading plugins (not at module import time)
        # to avoid polluting Python's import cache during test collection
        if PLUGIN_DIRECTORY not in sys.path:
            sys.path.append(PLUGIN_DIRECTORY)

        candidates = os.listdir(PLUGIN_DIRECTORY)

        # convert to Paths
        plugin_paths = [pathlib.Path(os.path.join(PLUGIN_DIRECTORY, name)) for name in candidates]

    # get all directories under the plugin directory
    plugin_paths = [path for path in plugin_paths if path.is_dir()]

    # load or reload each plugin
    for plugin_path in plugin_paths:
        load_or_reload_plugin(plugin_path)

    # if a plugin has been uninstalled/disabled remove it from LOADED_PLUGINS
    for name, plugin in LOADED_PLUGINS.copy().items():
        if not plugin["active"]:
            del LOADED_PLUGINS[name]

    refresh_event_type_map()


@measured
def load_plugin(path: pathlib.Path) -> None:
    """Load a plugin from the specified path."""
    load_or_reload_plugin(path)
    refresh_event_type_map()


# NOTE: specified_plugin_paths powers the `canvas run-plugins` command
def main(specified_plugin_paths: list[str] | None = None) -> None:
    """Run the server and the synchronize_plugins loop."""
    port = "50051"

    executor = ThreadPoolExecutor(max_workers=settings.PLUGIN_RUNNER_MAX_WORKERS)
    server = grpc.server(
        thread_pool=executor,
        options=(
            # set max message lengths to 64mb
            ("grpc.max_receive_message_length", 64 * 1024 * 1024),
            ("grpc.max_send_message_length", 64 * 1024 * 1024),
        ),
    )
    server.add_insecure_port("127.0.0.1:" + port)

    add_PluginRunnerServicer_to_server(PluginRunner(), server)

    log.info(f"Starting server, listening on port {port}")

    # Only install plugins and start the synchronizer thread if the plugin runner was not started
    # from the CLI
    synchronizer_thread = threading.Thread(target=synchronize_plugins_and_report_errors)
    if specified_plugin_paths is None:
        install_plugins()
        STOP_SYNCHRONIZER.clear()
        synchronizer_thread.start()

    load_plugins(specified_plugin_paths)

    server.start()

    try:
        server.wait_for_termination()
    except KeyboardInterrupt:
        pass
    finally:
        executor.shutdown(wait=True, cancel_futures=True)
        if synchronizer_thread.is_alive():
            STOP_SYNCHRONIZER.set()
            synchronizer_thread.join()


if __name__ == "__main__":
    main()



from enum import Enum


class SplitMethod(Enum):
    """Methods for splitting documents.

    HIGH_PRECISION: More accurate splitting at the cost of higher processing time.
    LOW_LATENCY: Faster splitting with potentially lower accuracy.
    """

    HIGH_PRECISION = "high_precision"
    LOW_LATENCY = "low_latency"


__exports__ = ()



import factory

from canvas_sdk.v1.data import ClaimLineItem, ClaimLineItemDiagnosisCode


class ClaimLineItemFactory(factory.django.DjangoModelFactory[ClaimLineItem]):
    """Factory for creating a ClaimLineItem."""

    class Meta:
        model = ClaimLineItem

    from_date = factory.Faker("date", pattern="%Y-%m-%d")
    claim = factory.SubFactory("canvas_sdk.test_utils.factories.ClaimFactory")


class ClaimLineItemDiagnosisCodeFactory(
    factory.django.DjangoModelFactory[ClaimLineItemDiagnosisCode]
):
    """Factory for creating a ClaimLineItemDiagnosisCode."""

    class Meta:
        model = ClaimLineItemDiagnosisCode

    line_item = factory.SubFactory(ClaimLineItemFactory)
    code = "T1490"



from typing import Final


class S3:
    """S3 client constant
    - ALGORITHM: AWS signature algorithm used (AWS4-HMAC-SHA256)
    - SAFE_CHARACTERS: Characters that don't need URL encoding
    - REQUEST_TYPE: AWS request type
    - UNSIGNED_PAYLOAD: Constant for unsigned payload in presigned URLs.
    """

    SERVICE_NAME: Final = "s3"
    ALGORITHM: Final = "AWS4-HMAC-SHA256"
    SAFE_CHARACTERS: Final = "-._~"
    REQUEST_TYPE: Final = "aws4_request"
    UNSIGNED_PAYLOAD: Final = "UNSIGNED-PAYLOAD"


__exports__ = ()



from django.contrib.postgres.fields import ArrayField
from django.db import models

from canvas_sdk.v1.data.base import TimestampedModel


class ProtocolResultStatus(models.TextChoices):
    """Constants for ProtocolResultStatus."""

    STATUS_DUE = "due"
    STATUS_SATISFIED = "satisfied"
    STATUS_NOT_APPLICABLE = "not_applicable"
    STATUS_PENDING = "pending"
    STATUS_NOT_RELEVANT = "not_relevant"


class ProtocolResult(TimestampedModel):
    """ProtocolResult."""

    class Meta:
        abstract = True

    title = models.TextField(blank=True, default="")
    narrative = models.TextField(blank=True, default="")
    result_identifiers = ArrayField(base_field=models.TextField())
    types = ArrayField(base_field=models.TextField())
    protocol_key = models.TextField(default="")
    plugin_name = models.TextField(null=True, blank=True)
    status = models.TextField(
        choices=ProtocolResultStatus.choices, default=ProtocolResultStatus.STATUS_NOT_APPLICABLE
    )
    due_in = models.IntegerField(null=True, default=-1)
    days_of_notice = models.IntegerField(default=30)
    snoozed = models.BooleanField(default=False)
    sources = models.JSONField(default=dict)
    recommendations = models.JSONField(default=dict)
    top_recommendation_key = models.TextField(blank=True, default="")
    next_review = models.DateTimeField(null=True)
    feedback_enabled = models.BooleanField(default=False)
    plugin_can_be_snoozed = models.BooleanField(default=False)


__exports__ = (
    "ProtocolResult",
    "ProtocolResultStatus",
)



from typing import NamedTuple


class FileContent(NamedTuple):
    """Container for file content with MIME type information.

    Attributes:
        mime_type: The MIME type of the file content (e.g., "application/pdf", "image/png").
        content: The file content as bytes, typically base64-encoded for API transmission.
    """

    mime_type: str
    content: bytes
    size: int


__exports__ = ()



from django.db import models

from canvas_sdk.v1.data.base import IdentifiableModel, TimestampedModel
from canvas_sdk.v1.data.coding import Coding


class CareTeamMembershipStatus(models.TextChoices):
    """CareTeamMembershipStatus."""

    PROPOSED = "proposed", "Proposed"
    ACTIVE = "active", "Active"
    SUSPENDED = "suspended", "Suspended"
    INACTIVE = "inactive", "Inactive"
    ENTERED_IN_ERROR = "entered-in-error", "Entered in Error"


class CareTeamRole(Coding):
    """CareTeamRole."""

    class Meta:
        db_table = "canvas_sdk_data_api_careteamrole_001"

    active = models.BooleanField()

    def __str__(self) -> str:
        return self.display


class CareTeamMembership(TimestampedModel, IdentifiableModel):
    """CareTeamMembership."""

    class Meta:
        db_table = "canvas_sdk_data_api_careteammembership_001"

    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="care_team_memberships", null=True
    )
    staff = models.ForeignKey(
        "v1.Staff", on_delete=models.DO_NOTHING, related_name="care_team_memberships", null=True
    )
    role = models.ForeignKey(
        "v1.CareTeamRole", related_name="care_teams", on_delete=models.DO_NOTHING, null=True
    )
    status = models.CharField(choices=CareTeamMembershipStatus.choices, max_length=20)
    lead = models.BooleanField()
    role_code = models.CharField(max_length=255)
    role_system = models.CharField(max_length=255)
    role_display = models.CharField(max_length=255)

    def __str__(self) -> str:
        return f"id={self.id}"


__exports__ = (
    "CareTeamMembershipStatus",
    "CareTeamRole",
    "CareTeamMembership",
)



from __future__ import annotations

from datetime import datetime
from typing import NamedTuple


class S3Item(NamedTuple):
    """S3 object metadata.

    Attributes:
        key: Object key (path) in the S3 bucket
        size: Object size in bytes
        last_modified: Timestamp of the last modification
    """

    key: str
    size: int
    last_modified: datetime


__exports__ = ("S3Item",)



import json
import re
from base64 import b64encode
from collections.abc import Mapping, Sequence
from http import HTTPStatus
from typing import Any

from pydantic_core import InitErrorDetails

from canvas_sdk.effects import Effect, EffectType, _BaseEffect

JSON = Mapping[str, "JSON"] | Sequence["JSON"] | int | float | str | bool | None


class Response(_BaseEffect):
    """SimpleAPI response class."""

    content: bytes | None
    status_code: int
    headers: dict[str, Any] | None

    def __init__(
        self,
        content: bytes | None = None,
        status_code: HTTPStatus = HTTPStatus.OK,
        headers: Mapping[str, Any] | None = None,
        content_type: str | None = None,
    ) -> None:
        headers = {**(headers or {})}
        if content_type:
            headers["Content-Type"] = content_type

        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,  # type: ignore[call-arg]
        )

    def apply(self) -> Effect:
        """Convert the response into an effect."""
        payload = {
            "headers": self.headers or {},
            "body": b64encode(self.content or b"").decode(),
            "status_code": self.status_code,
        }

        return Effect(type=EffectType.SIMPLE_API_RESPONSE, payload=json.dumps(payload))


class JSONResponse(Response):
    """SimpleAPI JSON response class."""

    def __init__(
        self,
        content: JSON,
        status_code: HTTPStatus = HTTPStatus.OK,
        headers: Mapping[str, Any] | None = None,
    ):
        super().__init__(
            json.dumps(content).encode(), status_code, headers, content_type="application/json"
        )


class PlainTextResponse(Response):
    """SimpleAPI plain text response class."""

    def __init__(
        self,
        content: str,
        status_code: HTTPStatus = HTTPStatus.OK,
        headers: Mapping[str, Any] | None = None,
    ):
        super().__init__(content.encode(), status_code, headers, content_type="text/plain")


class HTMLResponse(Response):
    """SimpleAPI HTML response class."""

    def __init__(
        self,
        content: str,
        status_code: HTTPStatus = HTTPStatus.OK,
        headers: Mapping[str, Any] | None = None,
    ):
        super().__init__(content.encode(), status_code, headers, content_type="text/html")


class AcceptConnection(_BaseEffect):
    """AcceptConnection effect."""

    class Meta:
        effect_type = EffectType.SIMPLE_API_WEBSOCKET_ACCEPT


class DenyConnection(_BaseEffect):
    """DenyConnection effect."""

    message: str | None = None

    class Meta:
        effect_type = EffectType.SIMPLE_API_WEBSOCKET_DENY

    @property
    def values(self) -> dict[str, Any]:
        """Make the payload."""
        return {"message": self.message} if self.message else {}


class Broadcast(_BaseEffect):
    """Broadcast effect."""

    class Meta:
        effect_type = EffectType.SIMPLE_API_WEBSOCKET_BROADCAST

    channel: str
    message: dict[str, Any]

    @property
    def values(self) -> dict[str, Any]:
        """Make the payload."""
        return {"channel": self.channel, "message": self.message}

    def is_valid_channel_name(self) -> bool:
        """Check if the channel name is valid."""
        return re.fullmatch(r"[\w-]+", self.channel) is not None

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if not self.is_valid_channel_name():
            errors.append(
                self._create_error_detail(
                    "value",
                    "Invalid channel name. Channel name must be alphanumeric and can contain underscores or hyphens.",
                    self.channel,
                )
            )

        return errors


__exports__ = (
    "JSON",
    "Response",
    "JSONResponse",
    "PlainTextResponse",
    "HTMLResponse",
    "AcceptConnection",
    "DenyConnection",
    "Broadcast",
)



from canvas_sdk.effects.claim.payment.base import (
    ClaimAllocation,
    LineItemTransaction,
    PaymentMethod,
)
from canvas_sdk.effects.payment.post_claim_payment import PostClaimPayment

__all__ = __exports__ = (
    "PostClaimPayment",
    "ClaimAllocation",
    "LineItemTransaction",
    "PaymentMethod",
)



from enum import Enum


class ParserTarget(Enum):
    """Target output formats for document parsing.

    MARKDOWN: Parse the document to markdown format.
    SPATIAL: Parse the document preserving spatial layout information.
    """

    MARKDOWN = "markdown"
    SPATIAL = "spatial"


__exports__ = ()



from enum import Enum
from typing import Any, overload

from pydantic import BaseModel, ConfigDict, field_validator

from canvas_sdk.commands.base import _BaseCommand
from canvas_sdk.effects.base import EffectType, _BaseEffect


class Recommendation(BaseModel):
    """
    A Recommendation for a Protocol Card.
    """

    model_config = ConfigDict(strict=True, validate_assignment=True)

    title: str = ""
    button: str = ""
    href: str | None = None
    commands: list[Any] | None = None

    @field_validator("commands", mode="before")
    @classmethod
    def check_subclass(cls, commands: list[Any] | None) -> list[Any] | None:
        """Validates that all commands are subclasses of _BaseCommand or dicts."""
        if commands is None:
            return commands
        for command in commands:
            if not isinstance(command, (_BaseCommand, dict)):
                raise TypeError(f"'{type(command)}' must be subclass of _BaseCommand")
        return commands

    @property
    def values(self) -> dict:
        """The ProtocolCard recommendation's values."""
        return {
            "title": self.title,
            "button": self.button,
            "href": self.href,
            "commands": [
                command.recommendation_context() if isinstance(command, _BaseCommand) else command
                for command in self.commands
            ]
            if self.commands
            else [],
        }


class ProtocolCard(_BaseEffect):
    """
    An Effect that will result in a protocol card in Canvas.
    """

    class Status(Enum):
        DUE = "due"
        SATISFIED = "satisfied"
        NOT_APPLICABLE = "not_applicable"
        PENDING = "pending"
        NOT_RELEVANT = "not_relevant"

    class Meta:
        effect_type = EffectType.ADD_OR_UPDATE_PROTOCOL_CARD
        apply_required_fields = ("patient_id|patient_filter", "key")

    patient_id: str | None = None
    key: str | None = None
    title: str = ""
    narrative: str = ""
    recommendations: list[Recommendation] = []
    status: Status = Status.DUE
    feedback_enabled: bool = False
    due_in: int = -1
    can_be_snoozed: bool = False

    @property
    def values(self) -> dict[str, Any]:
        """The ProtocolCard's values."""
        return {
            "title": self.title,
            "narrative": self.narrative,
            "recommendations": [
                rec.values | {"key": i} for i, rec in enumerate(self.recommendations)
            ],
            "status": self.status.value,
            "feedback_enabled": self.feedback_enabled,
            "due_in": self.due_in,
            "can_be_snoozed": self.can_be_snoozed,
        }

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {
            "patient": self.patient_id,
            "patient_filter": self.patient_filter,
            "key": self.key,
            "data": self.values,
        }

    @overload
    def add_recommendation(
        self,
        title: str = "",
        button: str = "",
        href: str | None = None,
        command: str | None = None,
        context: dict | None = None,
    ) -> None: ...

    @overload
    def add_recommendation(
        self,
        title: str = "",
        button: str = "",
        href: str | None = None,
        commands: list[Any] | None = None,
    ) -> None: ...

    def add_recommendation(  # type: ignore[misc]
        self,
        title: str = "",
        button: str = "",
        href: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Adds a recommendation to the protocol card's list of recommendations."""
        if "commands" in kwargs and "command" in kwargs:
            raise ValueError("Cannot provide both 'commands' and 'command'")

        commands: list[Any] = []
        if "command" in kwargs:
            command = kwargs.pop("command")
            context = kwargs.pop("context", {}) or {}
            if command:
                commands.append({"command": {"type": command}, "context": context})
        elif "commands" in kwargs:
            commands = kwargs.pop("commands") or []

        recommendation = Recommendation(
            title=title,
            button=button,
            href=href,
            commands=commands,
        )
        self.recommendations.append(recommendation)


__exports__ = (
    "Recommendation",
    "ProtocolCard",
)



from datetime import date, datetime
from enum import Enum
from typing import Any
from uuid import UUID

from pydantic import BaseModel, ConfigDict
from pydantic_core import InitErrorDetails, PydanticCustomError, ValidationError


class Model(BaseModel):
    """A base model that includes validation methods."""

    class Meta:
        pass

    model_config = ConfigDict(
        strict=True,
        revalidate_instances="always",
        validate_assignment=True,
        json_encoders={
            date: lambda v: v.isoformat(),
            datetime: lambda v: v.isoformat(),
            Enum: lambda v: v.value,
        },
    )

    def _get_effect_method_required_fields(self, method: Any) -> tuple:
        return getattr(self.Meta, f"{method}_required_fields", ())

    def _create_error_detail(self, type: str, message: str, value: Any) -> InitErrorDetails:
        return InitErrorDetails({"type": PydanticCustomError(type, message), "input": value})

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        required_fields = self._get_effect_method_required_fields(method)
        class_name = self.__repr_name__()  # type: ignore[misc]

        class_name_article = "an" if class_name.startswith(("A", "E", "I", "O", "U")) else "a"

        error_details = []
        for field in required_fields:
            fields = field.split("|")
            if not all(getattr(self, f) is None for f in fields):
                continue
            field_description = " or ".join([f"'{f}'" for f in fields])
            message = f"Field {field_description} is required to {method.replace('_', ' ')} {class_name_article} {class_name}"
            error = self._create_error_detail("missing", message, None)
            error_details.append(error)

        return error_details

    def _validate_before_effect(self, method: str) -> None:
        self.model_validate(self)
        if error_details := self._get_error_details(method):
            raise ValidationError.from_exception_data(self.__class__.__name__, error_details)


class TrackableFieldsModel(Model):
    """
    A base model with additional functionality for tracking modified fields.

    Attributes:
        _dirty_keys (set[str]): A set to track which fields have been modified.
    """

    _dirty_excluded_keys: list[str] = [
        "note_uuid",
    ]

    _dirty_keys: set[str] = set()

    def __init__(self, /, **data: Any) -> None:
        """Initialize the command and mark all provided keys as dirty."""
        super().__init__(**data)

        # Initialize a set to track which fields have been modified.
        self._dirty_keys = set()

        # Explicitly mark all keys provided in the constructor as dirty.
        self._dirty_keys.update(data.keys())

    def __setattr__(self, name: str, value: Any) -> None:
        """Set an attribute and mark it as dirty unless excluded."""
        if not name.startswith("_") and name not in self._dirty_excluded_keys:
            self._dirty_keys.add(name)
        super().__setattr__(name, value)

    def is_dirty(self, key: str) -> bool:
        """Returns True if the given property has been modified (i.e. marked as dirty), False otherwise."""
        return key in self._dirty_keys

    @property
    def values(self) -> dict:
        """Return a dictionary of modified attributes with type-specific transformations."""
        result = {}
        for key in self._dirty_keys:
            value = getattr(self, key)
            if isinstance(value, Enum):
                # If it's an enum, use its .value.
                result[key] = value.value if value else None
            elif isinstance(value, date | datetime):
                # If it's a date/datetime, use isoformat().
                result[key] = value.isoformat() if value else None
            elif isinstance(value, UUID):
                # If it's a UUID, use its string representation.
                result[key] = str(value) if value else None
            else:
                # For strings, integers, or any other type, return as is.
                result[key] = value
        return result



from dataclasses import dataclass
from typing import Self

from canvas_sdk.clients.extend_ai.constants.split_method import SplitMethod
from canvas_sdk.clients.extend_ai.structures.config.part_range import PartRange
from canvas_sdk.clients.extend_ai.structures.structure import Structure


@dataclass(frozen=True)
class AdvancedOptionsSplitter(Structure):
    """Advanced configuration options for document splitting processors.

    Attributes:
        split_identifier_rules: Custom rules for identifying split points.
        split_method: The method to use for splitting (high precision or standard).
        page_ranges: List of page ranges to process (empty list means all pages).
    """

    split_identifier_rules: str
    split_method: SplitMethod
    page_ranges: list[PartRange]

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create an AdvancedOptionsSplitter instance from a dictionary.

        Args:
            data: Dictionary containing advanced splitter options from the API.

        Returns:
            A new AdvancedOptionsSplitter instance with defaults for missing values.
        """
        return cls(
            split_identifier_rules=data.get("splitIdentifierRules") or "",
            split_method=SplitMethod(data.get("splitMethod") or SplitMethod.HIGH_PRECISION),
            page_ranges=[PartRange.from_dict(item) for item in data.get("page_ranges") or []],
        )

    def to_dict(self) -> dict:
        """Convert this AdvancedOptionsSplitter to a dictionary.

        Returns:
            Dictionary representation suitable for API requests.
        """
        return {
            "splitIdentifierRules": self.split_identifier_rules,
            "splitMethod": self.split_method.value,
            "pageRanges": [item.to_dict() for item in self.page_ranges],
        }


__exports__ = ()



from dataclasses import dataclass

from canvas_sdk.clients.llms.structures.settings.llm_settings import LlmSettings


@dataclass
class LlmSettingsAnthropic(LlmSettings):
    """Configuration settings for Anthropic Claude LLM API.

    Extends LlmSettings with Anthropic-specific parameters.

    Attributes:
        api_key: API authentication key for the LLM service (inherited).
        model: Name or identifier of the LLM model to use (inherited).
        temperature: Controls randomness in responses (0.0-1.0).
        max_tokens: Maximum number of tokens to generate.
    example:
        ```python3
        LlmSettingsAnthropic(
            api_key=environ.get("anthropic_key"),
            model="claude-sonnet-4-5-20250929",
            temperature=0.78,
            max_tokens=8192,
        )
        ```
    """

    temperature: float
    max_tokens: float

    def to_dict(self) -> dict:
        """Convert settings to Anthropic API request format.

        Returns:
            Dictionary containing model name, temperature, and max_tokens.
        """
        return super().to_dict() | {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
        }


__exports__ = ("LlmSettingsAnthropic",)



from dataclasses import dataclass
from typing import Self

from canvas_sdk.clients.sendgrid.constants.status_email import StatusEmail
from canvas_sdk.clients.sendgrid.structures.email_event import EmailEvent
from canvas_sdk.clients.sendgrid.structures.structure import Structure


@dataclass(frozen=True)
class SentEmailDetail(Structure):
    """Represents a sent email with its current event history."""

    from_email: str
    message_id: str
    subject: str
    to_email: str
    status: StatusEmail
    events: list[EmailEvent]

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create SentEmailDetail instance from dictionary."""
        return cls(
            from_email=data["from_email"],
            message_id=data["sg_message_id"],
            subject=data["subject"],
            to_email=data["to_email"],
            status=StatusEmail(data["status"]),
            events=[EmailEvent.from_dict(e) for e in data["events"]],
        )

    def to_dict(self) -> dict:
        """Convert sent email detailed to dictionary format."""
        return {
            "fromEmail": self.from_email,
            "messageId": self.message_id,
            "subject": self.subject,
            "toEmail": self.to_email,
            "status": self.status.value,
            "events": [e.to_dict() for e in self.events],
        }


__exports__ = ()



from canvas_sdk.clients.sendgrid.structures.address import Address
from canvas_sdk.clients.sendgrid.structures.attachment import Attachment
from canvas_sdk.clients.sendgrid.structures.body_content import BodyContent
from canvas_sdk.clients.sendgrid.structures.criterion_datetime import CriterionDatetime
from canvas_sdk.clients.sendgrid.structures.email import Email
from canvas_sdk.clients.sendgrid.structures.email_event import EmailEvent
from canvas_sdk.clients.sendgrid.structures.event_webhook import EventWebhook
from canvas_sdk.clients.sendgrid.structures.event_webhook_record import EventWebhookRecord
from canvas_sdk.clients.sendgrid.structures.logged_email_criteria import LoggedEmailCriteria
from canvas_sdk.clients.sendgrid.structures.parse_setting import ParseSetting
from canvas_sdk.clients.sendgrid.structures.parsed_attachment import ParsedAttachment
from canvas_sdk.clients.sendgrid.structures.parsed_email import ParsedEmail
from canvas_sdk.clients.sendgrid.structures.parsed_envelope import ParsedEnvelope
from canvas_sdk.clients.sendgrid.structures.parsed_header import ParsedHeader
from canvas_sdk.clients.sendgrid.structures.recipient import Recipient
from canvas_sdk.clients.sendgrid.structures.request_failed import RequestFailed
from canvas_sdk.clients.sendgrid.structures.sent_email import SentEmail
from canvas_sdk.clients.sendgrid.structures.sent_email_detail import SentEmailDetail
from canvas_sdk.clients.sendgrid.structures.settings import Settings

__all__ = __exports__ = (
    "Address",
    "Attachment",
    "BodyContent",
    "CriterionDatetime",
    "Email",
    "EmailEvent",
    "EventWebhook",
    "EventWebhookRecord",
    "LoggedEmailCriteria",
    "ParseSetting",
    "ParsedAttachment",
    "ParsedEmail",
    "ParsedEnvelope",
    "ParsedHeader",
    "Recipient",
    "RequestFailed",
    "SentEmail",
    "SentEmailDetail",
    "Settings",
)



from django.db import models

from canvas_sdk.v1.data.base import AuditedModel, IdentifiableModel
from canvas_sdk.v1.data.coding import Coding


class DetectedIssue(AuditedModel, IdentifiableModel):
    """DetectedIssue."""

    class Meta:
        db_table = "canvas_sdk_data_api_detectedissue_001"

    class Status(models.TextChoices):
        REGISTERED = "registered", "Registered"
        PRELIMINARY = "preliminary", "Preliminary"
        CANCELLED = "cancelled", "Cancelled"
        AMENDED = "amended", "Amended"
        FINAL = "final", "Final"
        CORRECTED = "corrected", "Corrected"
        ENTERED_IN_ERROR = "entered-in-error", "Entered in Error"

    class Severity(models.TextChoices):
        HIGH = "high", "High"
        MODERATE = "moderate", "Moderate"
        LOW = "low", "Low"

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    identified = models.DateTimeField()
    patient = models.ForeignKey(
        "v1.Patient", on_delete=models.DO_NOTHING, related_name="detected_issues", null=True
    )
    code = models.CharField(max_length=20)
    status = models.CharField(max_length=16, choices=Status.choices)
    severity = models.CharField(max_length=10, choices=Severity.choices, blank=True, default="")
    reference = models.CharField(max_length=200)
    issue_identifier = models.CharField(max_length=255)
    issue_identifier_system = models.CharField(max_length=255)
    detail = models.TextField()


class DetectedIssueEvidence(IdentifiableModel, Coding):
    """DetectedIssueEvidence."""

    class Meta:
        db_table = "canvas_sdk_data_api_detectedissueevidence_001"

    detected_issue = models.ForeignKey(
        DetectedIssue,
        on_delete=models.DO_NOTHING,
        related_name="evidence",
        null=True,
    )


__exports__ = ("DetectedIssue", "DetectedIssueEvidence")



from dataclasses import dataclass
from datetime import datetime
from typing import Self

from canvas_sdk.clients.extend_ai.constants.processor_type import ProcessorType
from canvas_sdk.clients.extend_ai.structures.config.config_classification import (
    ConfigClassification,
)
from canvas_sdk.clients.extend_ai.structures.config.config_extraction import ConfigExtraction
from canvas_sdk.clients.extend_ai.structures.config.config_splitter import ConfigSplitter
from canvas_sdk.clients.extend_ai.structures.processor_meta import ProcessorMeta
from canvas_sdk.clients.extend_ai.structures.structure import Structure


@dataclass(frozen=True)
class ProcessorVersion(Structure):
    """Represents a specific version of an Extend AI processor.

    Attributes:
        id: The unique identifier for this processor version.
        version: The version name (e.g., "draft", "v1", "published").
        description: A description of this processor version.
        processor: Metadata about the processor this version belongs to.
        config: The configuration for this processor version.
        created_at: The timestamp when this version was created.
        updated_at: The timestamp when this version was last updated.
    """

    id: str
    version: str
    description: str
    processor: ProcessorMeta
    config: ConfigClassification | ConfigExtraction | ConfigSplitter
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a ProcessorVersion instance from a dictionary.

        Args:
            data: Dictionary containing processor version data from the API.

        Returns:
            A new ProcessorVersion instance with the appropriate config type.
        """
        processor_type = ProcessorType(data["processorType"])
        config: ConfigClassification | ConfigExtraction | ConfigSplitter
        if processor_type == ProcessorType.EXTRACT:
            config = ConfigExtraction.from_dict(data["config"])
        elif processor_type == ProcessorType.CLASSIFY:
            config = ConfigClassification.from_dict(data["config"])
        else:  # processor_type==ExtendProcessorType.SPLITTER
            config = ConfigSplitter.from_dict(data["config"])

        return cls(
            id=data["id"],
            version=data["version"],
            description=data.get("description") or "",
            processor=ProcessorMeta(
                id=data["processorId"],
                name=data["processorName"],
                type=processor_type,
                created_at=None,
                updated_at=None,
            ),
            config=config,
            created_at=datetime.fromisoformat(data["createdAt"]),
            updated_at=datetime.fromisoformat(data["updatedAt"]),
        )

    def to_dict(self) -> dict:
        """Convert this ProcessorVersion to a dictionary.

        Returns:
            Dictionary representation suitable for API requests.
        """
        return {
            "id": self.id,
            "version": self.version,
            "description": self.description,
            "processor": self.processor.to_dict(),
            "config": self.config.to_dict(),
            "createdAt": self.created_at.isoformat() if self.created_at else None,
            "updatedAt": self.updated_at.isoformat() if self.updated_at else None,
        }


__exports__ = ()



from enum import Enum


class ProcessorType(Enum):
    """Types of processors available in Extend AI.

    EXTRACT: Extracts structured data from documents based on a schema.
    CLASSIFY: Classifies documents into predefined categories.
    SPLITTER: Splits documents into separate sections or parts.
    """

    EXTRACT = "EXTRACT"
    CLASSIFY = "CLASSIFY"
    SPLITTER = "SPLITTER"


__exports__ = ()



import factory

from canvas_sdk.v1.data import NoteTask, Task, TaskComment, TaskLabel, TaskMetadata, TaskTaskLabel
from canvas_sdk.v1.data.common import ColorEnum, Origin
from canvas_sdk.v1.data.task import TaskLabelModule, TaskStatus, TaskType


class TaskFactory(factory.django.DjangoModelFactory[Task]):
    """Factory for creating Task."""

    class Meta:
        model = Task

    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    creator = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    assignee = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    task_type = TaskType.TASK
    tag = factory.Faker("word")
    title = factory.Faker("sentence", nb_words=4)
    status = TaskStatus.OPEN


class TaskMetadataFactory(factory.django.DjangoModelFactory[TaskMetadata]):
    """Factory for creating TaskMetadata."""

    class Meta:
        model = TaskMetadata

    task = factory.SubFactory(TaskFactory)
    key = factory.Faker("word")
    value = factory.Faker("word")


class TaskCommentFactory(factory.django.DjangoModelFactory[TaskComment]):
    """Factory for creating TaskComment."""

    class Meta:
        model = TaskComment

    creator = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    task = factory.SubFactory(TaskFactory)
    body = factory.Faker("sentence", nb_words=10)


class TaskLabelFactory(factory.django.DjangoModelFactory[TaskLabel]):
    """Factory for creating TaskLabel."""

    class Meta:
        model = TaskLabel

    position = factory.Sequence(lambda n: n)
    color = ColorEnum.BLUE
    task_association = [Origin.REFERAL]
    name = factory.Faker("word")
    active = True
    modules = [TaskLabelModule.TASKS]


class TaskTaskLabelFactory(factory.django.DjangoModelFactory[TaskTaskLabel]):
    """Factory for creating TaskTaskLabel."""

    class Meta:
        model = TaskTaskLabel

    task = factory.SubFactory(TaskFactory)
    task_label = factory.SubFactory(TaskLabelFactory)


class NoteTaskFactory(factory.django.DjangoModelFactory[NoteTask]):
    """Factory for creating NoteTask."""

    class Meta:
        model = NoteTask

    note = factory.SubFactory("canvas_sdk.test_utils.factories.NoteFactory")
    task = factory.SubFactory(TaskFactory)
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    original_title = factory.Faker("sentence", nb_words=4)
    original_assignee = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    original_team = None
    original_role = None
    original_due = factory.Faker("date_time")
    internal_comment = factory.Faker("sentence", nb_words=6)
    originator = factory.SubFactory("canvas_sdk.test_utils.factories.CanvasUserFactory")
    committer = factory.SubFactory("canvas_sdk.test_utils.factories.CanvasUserFactory")



from dataclasses import dataclass

from canvas_sdk.clients.llms.structures.settings.llm_settings import LlmSettings


@dataclass
class LlmSettingsGpt4(LlmSettings):
    """Configuration settings for OpenAI GPT-4 LLM API.

    Extends LlmSettings with GPT-4-specific parameters.

    Attributes:
        api_key: API authentication key for the LLM service (inherited).
        model: Name or identifier of the LLM model to use (inherited).
        temperature: Controls randomness in responses (0.0-1.0).

    Example:
        ```python3
        LlmSettingsGpt4(
            api_key=environ.get("openai_key"),
            model="gpt-4o",
            temperature=2.0,
        )
        ```
    """

    temperature: float

    def to_dict(self) -> dict:
        """Convert settings to OpenAI GPT-4 API request format.

        Returns:
            Dictionary containing model name and temperature.
        """
        return super().to_dict() | {
            "temperature": self.temperature,
        }


__exports__ = ("LlmSettingsGpt4",)



class RequestFailed(RuntimeError):
    """Exception raised when a SendGrid API request fails."""

    def __init__(self, status_code: int, message: str):
        super().__init__(message)
        self.status_code = status_code
        self.message = message


__exports__ = ()



from dataclasses import dataclass


@dataclass(frozen=True)
class BodyContent:
    """Represents the body content of an email with MIME type."""

    type: str  # mime type
    value: str

    def to_dict(self) -> dict:
        """Convert body content to dictionary format."""
        return {
            "type": self.type,
            "value": self.value,
        }


__exports__ = ()



from decimal import Decimal
from typing import Any

from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType
from canvas_sdk.effects.claim.payment.base import (
    ClaimAllocation,
    PostPaymentBase,
)


class _PostClaimPayment(PostPaymentBase):
    """
    An Effect that posts a coverage or patient payment to a claim.
    """

    class Meta:
        effect_type = EffectType.POST_CLAIM_PAYMENT

    claim: ClaimAllocation

    @property
    def total_collected(self) -> str:
        """The total amount collected for this payment, calculated as sum of payments from line item transactions."""
        total = sum((item.payment or Decimal(0)) for item in self.claim.line_item_transactions)
        return str(total)

    @property
    def payment_collection_values(self) -> dict[str, Any]:
        """The values for the payment collection part of the payload."""
        base_values = super().payment_collection_values
        return base_values | {"total_collected": self.total_collected}

    @property
    def values(self) -> dict[str, Any]:
        """The values for the payload."""
        return {
            "payment_collection": self.payment_collection_values,
            "claims_allocation": [self.claim.to_dict()],
        }

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        errors.extend([self._create_error_detail(*e) for e in self.claim.validate()])
        errors.extend(self.validate_payment_method_fields())

        return errors


__exports__ = ()



from typing import Any
from uuid import UUID

from pydantic import constr
from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.v1.data import Claim, ClaimQueue


class _MoveClaimToQueue(_BaseEffect):
    """
    An Effect that moves a Claim to a Queue.
    """

    class Meta:
        effect_type = EffectType.MOVE_CLAIM_TO_QUEUE

    claim_id: UUID | str
    queue: constr(min_length=1, strip_whitespace=True)  # type: ignore[valid-type]

    @property
    def values(self) -> dict[str, Any]:
        """The claim_id and queue_id."""
        return {"claim_id": str(self.claim_id), "queue": self.queue}

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if not Claim.objects.filter(id=self.claim_id).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    "Claim does not exist",
                    self.claim_id,
                )
            )
        if not ClaimQueue.objects.filter(name=self.queue).exists():
            errors.append(
                self._create_error_detail(
                    "value",
                    "Queue does not exist",
                    self.queue,
                )
            )
        return errors


__exports__ = ()



# To run the tests, use the command `pytest` in the terminal or uv run pytest.
# Each test is wrapped inside a transaction that is rolled back at the end of the test.
# If you want to modify which files are used for testing, check the [tool.pytest.ini_options] section in pyproject.toml.
# For more information on testing Canvas plugins, see: https://docs.canvasmedical.com/sdk/testing-utils/

from unittest.mock import Mock

from canvas_sdk.effects import EffectType
from canvas_sdk.events import EventType
from canvas_sdk.test_utils.factories import PatientFactory
from canvas_sdk.v1.data.discount import Discount

from {{ cookiecutter.__package_name }}.handlers.event_handlers import NewOfficeVisitNoteHandler


# Test the handler's compute method with mocked event data
def test_handler_responds_to_new_office_visit_note() -> None:
    """Test that the handler originates commands for a new office visit note."""
    # Create a mock event with the expected structure
    mock_event = Mock()
    mock_event.type = EventType.NOTE_STATE_CHANGE_EVENT_CREATED
    mock_event.context = {
        "state": "NEW",
        "note_id": 123,
        "patient_id": 456,
    }

    # Mock the note instance
    mock_note = Mock()
    mock_note.uuid = "test-note-uuid-123"
    mock_note.note_type_version.name = "OFFICE VISIT"
    mock_event.target.instance = mock_note

    # Instantiate the handler with the mock event
    handler = NewOfficeVisitNoteHandler(event=mock_event)

    # Call compute and get the effects
    effects = handler.compute()

    # Assert that two effects were returned (vitals and goal)
    assert len(effects) == 2

    # Assert both effects have the correct type
    assert all(effect.type == EffectType.ORIGINATE_COMMAND for effect in effects)

    # Verify that the effects contain the correct note_uuid in their payloads
    for effect in effects:
        assert "test-note-uuid-123" in effect.payload


def test_handler_skips_non_new_notes() -> None:
    """Test that the handler skips notes that are not in NEW state."""
    mock_event = Mock()
    mock_event.type = EventType.NOTE_STATE_CHANGE_EVENT_CREATED
    mock_event.context = {
        "state": "SIGNED",
        "note_id": 123,
        "patient_id": 456,
    }

    # Mock the note instance
    mock_note = Mock()
    mock_note.uuid = "test-note-uuid-123"
    mock_note.note_type_version.name = "OFFICE VISIT"
    mock_event.target.instance = mock_note

    handler = NewOfficeVisitNoteHandler(event=mock_event)
    effects = handler.compute()

    # Assert that no effects were returned
    assert len(effects) == 0


def test_handler_skips_non_office_visit_notes() -> None:
    """Test that the handler skips notes that are not office visits."""
    mock_event = Mock()
    mock_event.type = EventType.NOTE_STATE_CHANGE_EVENT_CREATED
    mock_event.context = {
        "state": "NEW",
        "note_id": 123,
        "patient_id": 456,
    }

    # Mock the note instance
    mock_note = Mock()
    mock_note.uuid = "test-note-uuid-123"
    mock_note.note_type_version.name = "PROGRESS NOTE"
    mock_event.target.instance = mock_note

    handler = NewOfficeVisitNoteHandler(event=mock_event)
    effects = handler.compute()

    # Assert that no effects were returned
    assert len(effects) == 0


# Test that the handler class has the correct event type configured
def test_handler_event_configuration() -> None:
    """Test that the handler is configured to respond to the correct event type."""
    assert EventType.Name(EventType.NOTE_STATE_CHANGE_EVENT_CREATED) == NewOfficeVisitNoteHandler.RESPONDS_TO


# Example: You can use a factory to create a patient instance for testing purposes.
def test_factory_example() -> None:
    """Test that a patient can be created using the PatientFactory."""
    patient = PatientFactory.create()
    assert patient.id is not None


# Example: If a factory is not available, you can create an instance manually with the data model directly.
def test_model_example() -> None:
    """Test that a Discount instance can be created."""
    Discount.objects.create(name="10%", adjustment_group="30", adjustment_code="CO", discount=0.10)
    assert Discount.objects.first().pk is not None



class SimpleAPIException(Exception):
    """Base class for all SimpleAPI exceptions."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class AuthenticationError(SimpleAPIException):
    """Base class for all authentication exceptions."""

    pass


class NoAuthorizationHeaderError(AuthenticationError):
    """Exception class for requests without an Authorization header."""

    def __init__(self) -> None:
        super().__init__("Request has no Authorization header")


class AuthenticationSchemeError(AuthenticationError):
    """Exception class for requests that do not have a valid authentication scheme."""

    def __init__(self) -> None:
        super().__init__("Authorization header has no recognized authentication scheme")


class InvalidCredentialsFormatError(AuthenticationError):
    """Exception class for requests that have incorrectly-formatted credentials."""

    def __init__(self) -> None:
        super().__init__("Provided credentials are incorrectly formatted")


class InvalidCredentialsError(AuthenticationError):
    """Exception class for requests that have invalid credentials."""

    def __init__(self) -> None:
        super().__init__("Provided credentials are invalid")


__exports__ = (
    "SimpleAPIException",
    "AuthenticationError",
    "NoAuthorizationHeaderError",
    "AuthenticationSchemeError",
    "InvalidCredentialsFormatError",
    "InvalidCredentialsError",
)



from dataclasses import dataclass

from canvas_sdk.clients.twilio.structures import Capabilities
from canvas_sdk.clients.twilio.structures.request_failed import RequestFailed


@dataclass(frozen=True)
class SmsMms:
    """Represents an SMS or MMS message to be sent via Twilio.

    Attributes:
        number_from: The sender's phone number.
        number_from_sid: The Twilio SID of the sender's phone number.
        number_to: The recipient's phone number.
        message: The text content of the message.
        media_url: URL of media to attach (for MMS).
        status_callback_url: URL to receive delivery status updates.
    """

    number_from: str
    number_from_sid: str
    number_to: str
    message: str
    media_url: str
    status_callback_url: str

    def to_dict(self) -> dict:
        """Convert the SmsMms to a dictionary representation.

        Returns:
            Dictionary containing all message data.
        """
        return {
            "number_from": self.number_from,
            "number_from_sid": self.number_from_sid,
            "number_to": self.number_to,
            "message": self.message,
            "media_url": self.media_url,
            "status_callback_url": self.status_callback_url,
        }

    def to_api(self, capabilities: Capabilities) -> dict:
        """Convert the SmsMms to Twilio API request format with capability validation.

        Args:
            capabilities: Phone number capabilities to validate against.

        Returns:
            Dictionary formatted for Twilio API request.

        Raises:
            RequestFailed: If the phone number lacks required capabilities or no content is provided.
        """
        result = {
            "To": self.number_to,
            "From": self.number_from,
        }
        if self.media_url:
            if not capabilities.mms:
                raise RequestFailed(0, f"{self.number_from} cannot sent MMS")
            result["MediaUrl"] = self.media_url
        if self.message:
            if not capabilities.sms:
                raise RequestFailed(0, f"{self.number_from} cannot sent SMS")
            result["Body"] = self.message
        if self.status_callback_url:
            result["StatusCallback"] = self.status_callback_url
        if not ("Body" in result or "MediaUrl" in result):
            raise RequestFailed(0, "no content to be sent")
        return result


__exports__ = ()



from dataclasses import dataclass
from typing import Self

from canvas_sdk.clients.extend_ai.structures.structure import Structure


@dataclass(frozen=True)
class ResultExtraction(Structure):
    """Result of an extraction processor execution.

    Attributes:
        value: Dictionary containing the extracted data fields and their values.
    """

    value: dict

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a ResultExtraction instance from a dictionary.

        Args:
            data: Dictionary containing extraction result data from the API.

        Returns:
            A new ResultExtraction instance.
        """
        return cls(value=data["value"])

    def to_dict(self) -> dict:
        """Convert this ResultExtraction to a dictionary.

        Returns:
            Dictionary representation suitable for API requests.
        """
        return {"value": self.value}


__exports__ = ()



from abc import abstractmethod

import arrow
from cron_converter import Cron

from canvas_sdk.effects import Effect
from canvas_sdk.events import EventType
from canvas_sdk.handlers.base import BaseHandler


class CronTask(BaseHandler):
    """
    A type of handler that executes periodically according to a provided schedule.
    """

    RESPONDS_TO = EventType.Name(EventType.CRON)

    SCHEDULE: str = ""  # e.g. "* * * * *" for every minute.

    @abstractmethod
    def execute(self) -> list[Effect]:
        """
        Perform some work and return a list of effects.
        """

    def compute(self) -> list[Effect]:
        """
        See if the task should execute given the timestamp.
        """
        if not self.SCHEDULE:
            raise ValueError("You must set a SCHEDULE.")

        event_time = arrow.get(self.event.target.id).datetime
        cron = Cron(self.SCHEDULE)

        if event_time in cron:
            return self.execute()

        return []


__exports__ = ("CronTask",)



from enum import StrEnum
from typing import Any

from pydantic import Field
from pydantic_core import InitErrorDetails

from canvas_sdk.effects.base import EffectType, _BaseEffect
from canvas_sdk.v1.data import Application


class DefaultHomepageEffect(_BaseEffect):
    """
    An Effect that will set the default homepage for a user.
    """

    class Meta:
        effect_type = EffectType.HOMEPAGE_CONFIGURATION

    class Pages(StrEnum):
        PATIENTS = "/patients"
        SCHEDULE = "/schedule"
        REVENUE = "/revenue"
        CAMPAIGNS = "/campaigns"
        DATA_INTEGRATION = "/data-integration"

    page: Pages | None = Field(default=None, strict=False)
    application_identifier: str | None = None

    @property
    def values(self) -> dict[str, Any]:
        """Homepage configuration values."""
        return {
            "page": self.page.value if self.page else None,
            "application_identifier": self.application_identifier,
        }

    @property
    def effect_payload(self) -> dict[str, Any]:
        """The payload of the effect."""
        return {"data": self.values}

    def _get_error_details(self, method: Any) -> list[InitErrorDetails]:
        errors = super()._get_error_details(method)

        if self.page is None and self.application_identifier is None:
            errors.append(
                self._create_error_detail(
                    "value",
                    "Either page or application must be provided",
                    self.page,
                )
            )

        if self.application_identifier:
            application = Application.objects.filter(
                identifier=self.application_identifier
            ).exists()
            if not application:
                errors.append(
                    self._create_error_detail(
                        "application_identifier",
                        f"Application with identifier {self.application_identifier} does not exist",
                        self.application_identifier,
                    )
                )

        return errors


__exports__ = ("DefaultHomepageEffect",)



from dataclasses import dataclass


@dataclass(frozen=True)
class Address:
    """Represents an email address with name."""

    email: str
    name: str

    def to_dict(self) -> dict:
        """Convert address to dictionary format."""
        return {
            "email": self.email,
            "name": self.name,
        }


__exports__ = ()



import base64
import json
from http import HTTPStatus

from requests import exceptions

from canvas_sdk.clients.llms.constants.file_type import FileType
from canvas_sdk.clients.llms.libraries.llm_api import LlmApi
from canvas_sdk.clients.llms.structures.file_content import FileContent
from canvas_sdk.clients.llms.structures.llm_file_url import LlmFileUrl
from canvas_sdk.clients.llms.structures.llm_response import LlmResponse
from canvas_sdk.clients.llms.structures.llm_tokens import LlmTokens


class LlmAnthropic(LlmApi):
    """Anthropic Claude LLM API client.

    Implements the LlmBase interface for Anthropic's Claude API.
    """

    def _file_url_to_content_item(self, file_url: LlmFileUrl) -> dict | None:
        """Convert a file URL to an Anthropic content item."""
        if file_url.type == FileType.PDF:
            return {"type": "document", "source": {"type": "url", "url": file_url.url}}
        elif file_url.type == FileType.IMAGE:
            return {"type": "image", "source": {"type": "url", "url": file_url.url}}
        elif file_url.type == FileType.TEXT:
            content = self.str_content_of(file_url)
            return {
                "type": "document",
                "source": {
                    "type": "text",
                    "media_type": "text/plain",
                    "data": content,
                },
            }

    @classmethod
    def _file_content_to_content_item(cls, file_content: FileContent) -> dict | None:
        """Convert file content to an Anthropic content item."""
        if file_content.mime_type.startswith("image/"):
            item_type, source_type = "image", "base64"
        elif file_content.mime_type.endswith("/pdf"):
            item_type, source_type = "document", "base64"
        elif file_content.mime_type.startswith("text/"):
            return {
                "type": "document",
                "source": {
                    "type": "text",
                    "media_type": "text/plain",
                    "data": base64.standard_b64decode(file_content.content).decode("utf-8"),
                },
            }
        else:
            return None
        return {
            "type": item_type,
            "source": {
                "type": source_type,
                "media_type": file_content.mime_type,
                "data": file_content.content.decode("utf-8"),
            },
        }

    def to_dict(self) -> dict:
        """Convert prompts and add the necessary information to Anthropic API request format.

        Returns:
            Dictionary formatted for Anthropic API with messages array.
        """
        messages: list[dict] = []

        roles = {
            self.ROLE_SYSTEM: "user",
            self.ROLE_USER: "user",
            self.ROLE_MODEL: "assistant",
        }
        for prompt in self.prompts:
            role = roles[prompt.role]
            part = {"type": "text", "text": "\n".join(prompt.text)}
            # contiguous parts for the same role are merged
            if messages and messages[-1]["role"] == role:
                messages[-1]["content"].append(part)
            else:
                messages.append({"role": role, "content": [part]})

        # if there are files and the last message has the user's role
        if messages and messages[-1]["role"] == roles[self.ROLE_USER]:
            for file_url in self.file_urls:
                if item := self._file_url_to_content_item(file_url):
                    messages[-1]["content"].append(item)

            for file_content in self.file_contents:
                if item := self._file_content_to_content_item(file_content):
                    messages[-1]["content"].append(item)

            self.file_urls = []
            self.file_contents = []
        # structured output requested
        structured = {}
        if self.schema:
            name = self.schema.__name__
            structured = {
                "tool_choice": {"type": "tool", "name": name},
                "tools": [
                    {
                        "name": name,
                        # "description": "Provide the response using well-structured JSON.",
                        "input_schema": self.schema.model_json_schema(),
                    }
                ],
            }

        return self.settings.to_dict() | structured | {"messages": messages}

    @classmethod
    def _api_base_url(cls) -> str:
        return "https://api.anthropic.com"

    def request(self) -> LlmResponse:
        """Make a request to the Anthropic Claude API.

        Returns:
            Response containing status code, generated text, and token usage.
        """
        headers = {
            "Content-Type": "application/json",
            "anthropic-version": "2023-06-01",
            "x-api-key": self.settings.api_key,
        }
        data = json.dumps(self.to_dict())

        tokens = LlmTokens(prompt=0, generated=0)
        try:
            request = self.http.post("/v1/messages", headers=headers, data=data)
            code = request.status_code
            response = request.text
            if code == HTTPStatus.OK.value:
                content = json.loads(request.text)
                output = content.get("content", [{}])[0]
                if self.schema:
                    response = json.dumps(output.get("input", {}))
                else:
                    response = output.get("text", "")

                usage = content.get("usage", {})
                tokens = LlmTokens(
                    prompt=usage.get("input_tokens") or 0,
                    generated=usage.get("output_tokens") or 0,
                )
        except exceptions.RequestException as e:
            code = HTTPStatus.BAD_REQUEST
            response = f"Request failed: {e}"
            if message := getattr(e, "response", None):
                code = message.status_code
                response = message.text

        return LlmResponse(
            code=HTTPStatus(code),
            response=response,
            tokens=tokens,
        )


__exports__ = ("LlmAnthropic",)



from typing import Self

from django.db import models

from canvas_sdk.v1.data.report_template_base import (
    BaseReportTemplate,
    BaseReportTemplateField,
    BaseReportTemplateFieldOption,
    BaseReportTemplateQuerySet,
)


class SpecialtyReportTemplateQuerySet(BaseReportTemplateQuerySet):
    """QuerySet for SpecialtyReportTemplate with filtering methods."""

    def search(self, query: str) -> Self:
        """Perform full-text search using the search_keywords field."""
        return self.filter(search_keywords__icontains=query)

    def by_specialty(self, specialty_code: str) -> Self:
        """Filter templates by specialty taxonomy code."""
        return self.filter(specialty_code=specialty_code)


class SpecialtyReportTemplate(BaseReportTemplate):
    """Model for specialty report templates used for LLM-powered specialty/referral report parsing."""

    class Meta:
        db_table = "canvas_sdk_data_data_integration_specialtyreporttemplate_001"

    objects = models.Manager.from_queryset(SpecialtyReportTemplateQuerySet)()

    search_as = models.CharField(max_length=255, default="", blank=True)
    specialty_name = models.CharField(max_length=255, default="", blank=True)
    specialty_code = models.CharField(max_length=255, default="", blank=True)
    specialty_code_system = models.CharField(max_length=255, default="", blank=True)


class SpecialtyReportTemplateField(BaseReportTemplateField):
    """Model for field definitions within a specialty report template."""

    class Meta:
        db_table = "canvas_sdk_data_data_integration_specialtyrpttmplfield_001"

    report_template = models.ForeignKey(
        SpecialtyReportTemplate,
        on_delete=models.DO_NOTHING,
        related_name="fields",
    )


class SpecialtyReportTemplateFieldOption(BaseReportTemplateFieldOption):
    """Model for options for select/radio/checkbox fields in specialty report templates."""

    class Meta:
        db_table = "canvas_sdk_data_data_integration_specialtyrpttmplfieldopt_001"

    field = models.ForeignKey(
        SpecialtyReportTemplateField,
        on_delete=models.DO_NOTHING,
        related_name="options",
    )


__exports__ = (
    "SpecialtyReportTemplate",
    "SpecialtyReportTemplateField",
    "SpecialtyReportTemplateFieldOption",
)



from base64 import b64encode
from collections.abc import Iterator
from http import HTTPStatus
from typing import TypeVar
from urllib.parse import urlencode

from canvas_sdk.clients.twilio.constants import HttpMethod
from canvas_sdk.clients.twilio.constants.date_operation import DateOperation
from canvas_sdk.clients.twilio.structures import Phone
from canvas_sdk.clients.twilio.structures.media import Media
from canvas_sdk.clients.twilio.structures.message import Message
from canvas_sdk.clients.twilio.structures.request_failed import RequestFailed
from canvas_sdk.clients.twilio.structures.settings import Settings
from canvas_sdk.clients.twilio.structures.sms_mms import SmsMms
from canvas_sdk.clients.twilio.structures.structure import Structure
from canvas_sdk.utils import Http

Data = TypeVar("Data", bound=Structure)


class SmsClient:
    """Twilio API client for sending and retrieving SMS messages.

    Provides methods to interact with Twilio's messaging API including
    sending messages, retrieving message details, and handling callbacks.
    """

    def __init__(self, settings: Settings):
        """Initialize the Twilio client with authentication settings.

        Args:
            settings: Twilio account settings containing credentials.
        """
        self.settings = settings
        self.http = Http(f"https://api.twilio.com/2010-04-01/Accounts/{self.settings.account_sid}/")

    def _auth_header(self) -> dict:
        """Generate Basic authentication header for Twilio API requests.

        Returns:
            Dictionary containing the Authorization header with base64 encoded credentials.
        """
        auth_header = b64encode(f"{self.settings.key}:{self.settings.secret}".encode()).decode()
        return {
            "Authorization": f"Basic {auth_header}",
        }

    def _valid_content_list(self, url: str, key: str, returned: type[Data]) -> Iterator[Data]:
        """Fetch paginated list data from Twilio API.

        Args:
            url: The API endpoint URL to fetch from.
            key: The JSON key containing the list of items.
            returned: The Structure type to deserialize items into.

        Yields:
            Instances of the specified Structure type.

        Raises:
            RequestFailed: If any API request fails.
        """
        while True:
            request = self.http.get(url, headers=self._auth_header())
            if request.status_code != HTTPStatus.OK:
                raise RequestFailed(request.status_code, request.content.decode())
            response = request.json()
            for item in response[key]:
                yield returned.from_dict(item)
            url = response.get("next_page_uri")
            if not url:
                break

    def account_phone_numbers(self) -> Iterator[Phone]:
        """Retrieve all phone numbers associated with the Twilio account.

        Yields:
            Phone instances for each phone number in the account.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = "IncomingPhoneNumbers.json"
        yield from self._valid_content_list(url, "incoming_phone_numbers", Phone)

    def account_phone_number(self, phone_sid: str) -> Phone:
        """Retrieve details for a specific phone number by its SID.

        Args:
            phone_sid: The Twilio SID of the phone number.

        Returns:
            Phone instance with the phone number details.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"IncomingPhoneNumbers/{phone_sid}.json"
        request = self.http.get(url, headers=self._auth_header())
        if request.status_code != HTTPStatus.OK:
            raise RequestFailed(request.status_code, request.content.decode())
        return Phone.from_dict(request.json())

    def set_inbound_webhook(self, phone_sid: str, webhook_url: str, method: HttpMethod) -> bool:
        """Set the webhook URL for Twilio inbound messages.

        Args:
            phone_sid: The Twilio SID of the phone number.
            webhook_url: The URL used by Twilio to send the inbound messages.
            method: The HTTP method used by Twilio (GET or POST).

        Returns:
            True if the inbound webhook is correctly set.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"IncomingPhoneNumbers/{phone_sid}.json"
        data = {
            "SmsUrl": webhook_url,
            "SmsMethod": method.value,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"} | self._auth_header()
        request = self.http.post(url, data=data, headers=headers)
        if request.status_code != HTTPStatus.OK:
            raise RequestFailed(request.status_code, request.content.decode())
        return True

    def send_sms_mms(self, sms_mms: SmsMms) -> Message:
        """Send an SMS or MMS message via Twilio.

        Args:
            sms_mms: The message details to send.

        Returns:
            Message instance with the created message details.

        Raises:
            RequestFailed: If the API request fails or phone lacks required capabilities.
        """
        url = "Messages.json"
        phone = self.account_phone_number(sms_mms.number_from_sid)  # maybe to be cached?
        data = sms_mms.to_api(phone.capabilities)
        headers = {"Content-Type": "application/x-www-form-urlencoded"} | self._auth_header()
        request = self.http.post(url, data=data, headers=headers)
        if request.status_code != HTTPStatus.CREATED:
            raise RequestFailed(request.status_code, request.content.decode())
        return Message.from_dict(request.json())

    def retrieve_sms(self, message_id: str) -> Message:
        """Retrieve details for a specific SMS message by its ID.

        Args:
            message_id: The Twilio message SID to retrieve.

        Returns:
            Message instance with the message details.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"Messages/{message_id}.json"
        request = self.http.get(url, headers=self._auth_header())
        if request.status_code != HTTPStatus.OK:
            raise RequestFailed(request.status_code, request.content.decode())
        return Message.from_dict(request.json())

    def retrieve_all_sms(
        self,
        number_to: str,
        number_from: str,
        date_sent: str,
        date_operation: DateOperation,
    ) -> Iterator[Message]:
        """Retrieve all SMS messages matching the given filters.

        Args:
            number_to: Filter by recipient phone number (empty string for no filter).
            number_from: Filter by sender phone number (empty string for no filter).
            date_sent: Date to filter by in the format required by Twilio.
            date_operation: Date operation type (on, before, or after the date).

        Yields:
            Message instances matching the filter criteria.

        Raises:
            RequestFailed: If any API request fails.
        """
        url = "Messages.json"
        if params := self._all_sms_query_params(number_to, number_from, date_sent, date_operation):
            url = f"{url}?{params}"
        yield from self._valid_content_list(url, "messages", Message)

    @classmethod
    def _all_sms_query_params(
        cls,
        number_to: str,
        number_from: str,
        date_sent: str,
        date_operation: DateOperation,
    ) -> str:
        """Build URL-encoded query parameters for filtering SMS messages.

        Args:
            number_to: Filter by recipient phone number.
            number_from: Filter by sender phone number.
            date_sent: Date to filter by.
            date_operation: Date operation type (on, before, or after).

        Returns:
            URL-encoded query string with non-empty parameters.
        """
        date_field = {
            DateOperation.ON_EXACTLY: "DateSent",
            DateOperation.ON_AND_AFTER: "DateSentAfter",
            DateOperation.ON_AND_BEFORE: "DateSentBefore",
        }[date_operation]
        params = [
            (date_field, date_sent),
            ("To", number_to),
            ("From", number_from),
        ]
        return urlencode({k: v for k, v in params if v})

    def retrieve_media_list(self, message_id: str) -> Iterator[Media]:
        """Retrieve all media items attached to a message.

        Args:
            message_id: The Twilio message SID.

        Yields:
            Media instances for each media item attached to the message.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"Messages/{message_id}/Media.json"
        yield from self._valid_content_list(url, "media_list", Media)

    def retrieve_raw_media(self, message_id: str, media_sid: str) -> bytes:
        """Retrieve the raw binary content of a media item.

        Args:
            message_id: The Twilio message SID.
            media_sid: The Twilio media SID.

        Returns:
            Raw binary content of the media file.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"Messages/{message_id}/Media/{media_sid}"
        request = self.http.get(url, headers=self._auth_header())
        if request.status_code != HTTPStatus.OK:
            raise RequestFailed(request.status_code, request.content.decode())
        return request.content

    def delete_sms(self, message_id: str) -> bool:
        """Delete a message.

        Args:
            message_id: The Twilio message SID.

        Returns:
            True if the message was successfully deleted.

        Raises:
            RequestFailed: If the API request fails.
        """
        url = f"Messages/{message_id}.json"
        request = self.http.delete(url, headers=self._auth_header())
        if request.status_code != HTTPStatus.NO_CONTENT:
            raise RequestFailed(request.status_code, request.content.decode())
        return True


__exports__ = ()



from dataclasses import dataclass
from typing import Any, Self

from pydantic import Field

from canvas_sdk.effects import EffectType
from canvas_sdk.effects.base import _BaseEffect


@dataclass
class ValidationError:
    """
    Represents a validation error for an effect or command.

    Attributes:
        message: The validation error message to display
    """

    message: str

    def __post_init__(self) -> None:
        """
        Validate and normalize the error message after initialization.

        Raises:
            ValueError: If message is empty
        """
        if not self.message or not self.message.strip():
            raise ValueError("Error message cannot be empty")

        self.message = self.message.strip()

    def to_dict(self) -> dict[str, str]:
        """Convert the validation error to a dictionary."""
        return {"message": self.message}

    def __repr__(self) -> str:
        return f"ValidationError(message={self.message!r})"


class _BaseValidationErrorEffect(_BaseEffect):
    """
    Abstract effect to abort an event.
    """

    class Meta:
        effect_type = EffectType.UNKNOWN_EFFECT

    errors: list[ValidationError] = Field(default_factory=list)

    def add_error(self, message: str | ValidationError) -> Self:
        """
        Add a validation error to the effect.

        This method allows incremental building of validation errors
        and supports method chaining.

        Args:
            message: The error message to display, or a ValidationError object

        Returns:
            Self for method chaining

        Raises:
            ValueError: If message is empty

        Example:
            effect = CommandValidationErrorEffect()
            effect.add_error("Narrative is required")
            effect.add_error("Dosage must be positive")

            # Using ValidationError objects
            error = ValidationError("Already validated error")
            effect.add_error(error)

            # Method chaining
            effect.add_error("Error 1").add_error("Error 2")
        """
        if isinstance(message, ValidationError):
            self.errors.append(message)
        else:
            error = ValidationError(message=message)
            self.errors.append(error)
        return self

    @property
    def values(self) -> dict[str, Any]:
        """Payload to include in the Effect."""
        return {"errors": [error.to_dict() for error in self.errors]}


__exports__ = ("ValidationError",)



from enum import Enum


class CriterionOperation(Enum):
    """Comparison operations for SendGrid query criteria."""

    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    LOWER_THAN = "<"
    LOWER_THAN_OR_EQUAL = "<="
    EQUAL = "="


__exports__ = ()



from canvas_sdk.commands.commands.questionnaire import QuestionnaireCommand


class StructuredAssessmentCommand(QuestionnaireCommand):
    """A class for managing physical exam command."""

    class Meta:
        key = "structuredAssessment"


__exports__ = ("StructuredAssessmentCommand",)



from decimal import Decimal
from typing import Any, Self

from django.db import models
from django.db.models import Q, Sum
from django.db.models.functions import Coalesce

from canvas_sdk.v1.data.base import IdentifiableModel, TimestampedModel
from canvas_sdk.v1.data.note import PracticeLocationPOS


class ClaimLineItemStatus(models.TextChoices):
    """ClaimLineItemStatus."""

    ACTIVE = "active", "Active"
    REMOVED = "removed", "Removed"


class LineItemCodes(models.TextChoices):
    """LineItemCodes."""

    COPAY = "COPAY"
    UNLINKED = "UNLINKED"


class FamilyPlanningOptions(models.TextChoices):
    """FamilyPlanningOptions."""

    YES = ("Y", "Yes")
    NO = ("N", "No")


class ClaimLineItemQuerySet(models.QuerySet):
    """ClaimLineItemQuerySet."""

    def filter(self, *args: Any, **kwargs: Any) -> Self:
        """Apply standard filtering to the QuerySet."""
        return super().filter(*args, **kwargs)

    def exclude(self, *args: Any, **kwargs: Any) -> Self:
        """Apply standard exclusion to the QuerySet."""
        return super().exclude(*args, **kwargs)

    def apply_ordering(self) -> Self:
        """Apply ordering to the QuerySet."""
        return self.order_by("from_date", "-proc_code")

    def active(self) -> Self:
        """Return only active claim line items."""
        return self.exclude(status=ClaimLineItemStatus.REMOVED).apply_ordering()

    def exclude_removed_line_item_without_active_posting(self) -> Self:
        """Exclude out any line items that are deleted without any active postings."""
        return (
            self.filter(
                Q(status=ClaimLineItemStatus.ACTIVE)
                | Q(
                    newlineitempayments__isnull=False,
                    newlineitempayments__entered_in_error__isnull=True,
                )
                | Q(
                    newlineitemadjustments__isnull=False,
                    newlineitemadjustments__entered_in_error__isnull=True,
                )
                | Q(
                    lineitemtransfers__isnull=False,
                    lineitemtransfers__entered_in_error__isnull=True,
                )
            )
            .exclude_removed_line_items_without_balances()
            .apply_ordering()
            .distinct()
        )

    def exclude_removed_line_items_without_balances(self) -> Self:
        """Exclude removed line items with zero balances."""
        return (
            self.annotate(
                sum_of_all_payments=Coalesce(Sum("newlineitempayments__amount"), Decimal(0)),
                sum_of_all_adjustments=Coalesce(Sum("newlineitemadjustments__amount"), Decimal(0)),
                sum_of_all_transfers=Coalesce(Sum("lineitemtransfers__amount"), Decimal(0)),
            )
            .exclude(
                Q(
                    status=ClaimLineItemStatus.REMOVED,
                    sum_of_all_payments=0,
                    sum_of_all_adjustments=0,
                    sum_of_all_transfers=0,
                )
            )
            .distinct()
        )

    def exclude_copay_and_unlinked(self) -> Self:
        """Exclude both co-pay and unlinked line items."""
        return self.exclude(
            proc_code__in=[LineItemCodes.UNLINKED.value, LineItemCodes.COPAY.value]
        ).apply_ordering()

    def exclude_copay(self) -> Self:
        """Exclude only co-pay line items."""
        return self.exclude(proc_code=LineItemCodes.COPAY.value).apply_ordering()

    def exclude_unlinked(self) -> Self:
        """Exclude only unlinked line items."""
        return self.exclude(proc_code=LineItemCodes.UNLINKED.value).apply_ordering()


class ClaimLineItem(TimestampedModel, IdentifiableModel):
    """ClaimLineItem."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_claimlineitem_001"

    objects = ClaimLineItemQuerySet.as_manager()

    billing_line_item = models.ForeignKey("v1.BillingLineItem", on_delete=models.CASCADE, null=True)
    claim = models.ForeignKey("v1.Claim", on_delete=models.CASCADE, related_name="line_items")
    status = models.CharField(
        choices=ClaimLineItemStatus.choices, max_length=10, default=ClaimLineItemStatus.ACTIVE
    )
    charge = models.DecimalField(max_digits=8, decimal_places=2, default=0.00)
    from_date = models.CharField(max_length=10, default="0000-00-00")
    thru_date = models.CharField(max_length=10, default="0000-00-00")
    narrative = models.CharField(max_length=2000, default="", blank=True)
    ndc_code = models.CharField(max_length=100, default="", blank=True)
    ndc_dosage = models.CharField(max_length=100, default="", blank=True)
    ndc_measure = models.CharField(max_length=100, default="", blank=True)
    place_of_service = models.CharField(
        choices=PracticeLocationPOS.choices,
        max_length=2,
        default=PracticeLocationPOS.OFFICE,
        blank=True,
        null=True,
    )
    proc_code = models.CharField(max_length=10, default="")
    display = models.CharField(max_length=255, default="", blank=True)
    remote_chg_id = models.CharField(max_length=100, default="", blank=True)
    units = models.IntegerField(default=1)
    epsdt = models.CharField(max_length=2, default="", blank=True)
    family_planning = models.CharField(
        choices=FamilyPlanningOptions.choices, max_length=1, default="", blank=True
    )


class ClaimLineItemDiagnosisCode(TimestampedModel, IdentifiableModel):
    """ClaimLineItemDiagnosisCode."""

    class Meta:
        db_table = "canvas_sdk_data_quality_and_revenue_lineitemdiagnosiscode_001"
        ordering = ["claim_diagnosis_code__rank"]

    line_item = models.ForeignKey(
        "v1.ClaimLineItem", on_delete=models.CASCADE, related_name="diagnosis_codes"
    )
    claim_diagnosis_code = models.ForeignKey(
        "v1.ClaimDiagnosisCode",
        on_delete=models.CASCADE,
        related_name="line_item_diagnosis_codes",
        null=True,
    )
    code = models.CharField(max_length=20, blank=True)
    poa = models.CharField(max_length=1, default="", blank=True)
    linked = models.BooleanField(default=True)


__exports__ = (
    "ClaimLineItem",
    "ClaimLineItemStatus",
    "LineItemCodes",
    "FamilyPlanningOptions",
    "ClaimLineItemDiagnosisCode",
)



from django.db import close_old_connections, reset_queries


def thread_cleanup() -> None:
    """
    Call this method to do cleanup work at thread exit. For users doing
    advanced work with threads there is a need to do cleanup after each thread
    runs so that e.g. the database pool does not become exhausted.
    """
    reset_queries()
    close_old_connections()


__exports__ = ("thread_cleanup",)



from canvas_sdk.commands.commands.review.base import _BaseReview
from canvas_sdk.v1.data import UncategorizedClinicalDocument


class UncategorizedDocumentReviewCommand(_BaseReview):
    """A class for managing an Uncategorized Document Review command within a specific note."""

    class Meta:
        key = "uncategorizedDocumentReview"
        model = UncategorizedClinicalDocument


__exports__ = ("UncategorizedDocumentReviewCommand",)



from typing import Any

import factory
from django.utils import timezone

from canvas_sdk.v1.data import (
    LabOrder,
    LabOrderReason,
    LabOrderReasonCondition,
    LabPartner,
    LabPartnerTest,
    LabReport,
    LabReview,
    LabTest,
    LabValue,
    LabValueCoding,
)
from canvas_sdk.v1.data.common import (
    DocumentReviewMode,
    ReviewPatientCommunicationMethod,
    ReviewStatus,
)
from canvas_sdk.v1.data.lab import TransmissionType


class LabReportFactory(factory.django.DjangoModelFactory[LabReport]):
    """Factory for creating a LabReport."""

    class Meta:
        model = LabReport

    review_mode = DocumentReviewMode.REVIEW_REQUIRED
    junked = False
    requires_signature = False
    assigned_date = factory.LazyFunction(lambda: timezone.now())
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    transmission_type = TransmissionType.HL7
    for_test_only = False
    external_id = factory.Faker("uuid4")
    version = 1
    requisition_number = factory.Faker("bothify", text="REQ-########")
    review = None  # Optional field
    original_date = factory.LazyFunction(lambda: timezone.now())
    date_performed = factory.LazyFunction(lambda: timezone.now())
    custom_document_name = factory.Faker("text", max_nb_chars=500)


class LabReviewFactory(factory.django.DjangoModelFactory[LabReview]):
    """Factory for creating a LabReview."""

    class Meta:
        model = LabReview

    internal_comment = factory.Faker("paragraph")
    message_to_patient = factory.Faker("text", max_nb_chars=2048)
    status = ReviewStatus.STATUS_REVIEWING
    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    patient_communication_method = ReviewPatientCommunicationMethod.DELEGATED_CALL_CAN_LEAVE_MESSAGE


class LabValueFactory(factory.django.DjangoModelFactory[LabValue]):
    """Factory for creating a LabValue."""

    class Meta:
        model = LabValue

    report = factory.SubFactory(LabReportFactory)
    value = factory.Faker("bothify", text="##.#")
    units = factory.Faker("random_element", elements=["mg/dL", "mmol/L", "g/dL", "%", "U/L"])
    abnormal_flag = factory.Faker("random_element", elements=["", "H", "L", "HH", "LL"])
    reference_range = factory.Faker("bothify", text="#.#-##.#")
    low_threshold = factory.Faker("bothify", text="#.#")
    high_threshold = factory.Faker("bothify", text="##.#")
    comment = factory.Faker("sentence")
    observation_status = factory.Faker(
        "random_element", elements=["final", "preliminary", "corrected"]
    )


class LabValueCodingFactory(factory.django.DjangoModelFactory[LabValueCoding]):
    """Factory for creating a LabValueCoding."""

    class Meta:
        model = LabValueCoding

    value = factory.SubFactory(LabValueFactory)
    code = factory.Faker("bothify", text="#####-#")
    name = factory.Faker("text", max_nb_chars=256)
    system = factory.Faker("random_element", elements=["LOINC", "SNOMED", "CPT"])


class LabOrderFactory(factory.django.DjangoModelFactory[LabOrder]):
    """Factory for creating a LabOrder."""

    class Meta:
        model = LabOrder
        skip_postgeneration_save = True

    patient = factory.SubFactory("canvas_sdk.test_utils.factories.PatientFactory")
    ontology_lab_partner = factory.Faker("company")
    note = factory.SubFactory("canvas_sdk.test_utils.factories.NoteFactory")
    comment = factory.Faker("text", max_nb_chars=128)
    requisition_number = factory.Faker("bothify", text="REQ-########")
    is_patient_bill = False
    date_ordered = factory.LazyFunction(lambda: timezone.now())
    fasting_status = None  # Optional field
    specimen_collection_type = LabOrder.SpecimenCollectionType.ON_LOCATION
    transmission_type = TransmissionType.HL7
    courtesy_copy_type = None  # Optional field
    courtesy_copy_number = ""
    courtesy_copy_text = ""
    ordering_provider = factory.SubFactory("canvas_sdk.test_utils.factories.StaffFactory")
    parent_order = None  # Optional field
    healthgorilla_id = ""
    manual_processing_status = None  # Optional field
    manual_processing_comment = None  # Optional field
    labcorp_abn_url = ""

    @factory.post_generation
    def reports(self, create: Any, extracted: Any, **kwargs: Any) -> None:
        """Handle many-to-many relationship for reports."""
        if not create:
            return

        if extracted:
            for report in extracted:
                self.reports.add(report)


class LabOrderReasonFactory(factory.django.DjangoModelFactory[LabOrderReason]):
    """Factory for creating a LabOrderReason."""

    class Meta:
        model = LabOrderReason

    order = factory.SubFactory(LabOrderFactory)
    mode = LabOrderReason.LabReasonMode.MONITOR


class LabOrderReasonConditionFactory(factory.django.DjangoModelFactory[LabOrderReasonCondition]):
    """Factory for creating a LabOrderReasonCondition."""

    class Meta:
        model = LabOrderReasonCondition

    reason = factory.SubFactory(LabOrderReasonFactory)
    condition = None  # Optional field - would need a ConditionFactory


class LabTestFactory(factory.django.DjangoModelFactory[LabTest]):
    """Factory for creating a LabTest."""

    class Meta:
        model = LabTest

    ontology_test_name = factory.Faker("text", max_nb_chars=512)
    ontology_test_code = factory.Faker("bothify", text="####-#")
    status = LabTest.LabTestOrderStatus.NEW
    report = factory.SubFactory(LabReportFactory)
    aoe_code = ""
    procedure_class = ""
    specimen_type = factory.Faker("random_element", elements=["blood", "urine", "serum", "plasma"])
    specimen_source_code = factory.Faker("bothify", text="###")
    specimen_source_description = factory.Faker("text", max_nb_chars=255)
    specimen_source_coding_system = factory.Faker("random_element", elements=["LOINC", "SNOMED"])
    order = factory.SubFactory(LabOrderFactory)


class LabPartnerFactory(factory.django.DjangoModelFactory[LabPartner]):
    """Factory for creating a LabPartner."""

    class Meta:
        model = LabPartner

    name = factory.Faker("company")
    active = True
    electronic_ordering_enabled = True
    keywords = factory.Faker("words", nb=5)
    default_lab_account_number = factory.Faker("bothify", text="LAB-######")


class LabPartnerTestFactory(factory.django.DjangoModelFactory[LabPartnerTest]):
    """Factory for creating a LabPartnerTest."""

    class Meta:
        model = LabPartnerTest

    lab_partner = factory.SubFactory(LabPartnerFactory)
    order_code = factory.Faker("bothify", text="####")
    order_name = factory.Faker("text", max_nb_chars=256)
    keywords = factory.Faker("words", nb=5)
    cpt_code = factory.Faker("bothify", text="#####")
